from audioop import bias
import dgl
import torch
import torch.nn as nn
import torch.nn.functional as F
import dgl.nn.pytorch as dglnn

from . import BaseModel, register_model


@register_model('ieHGCN')
class ieHGCN(BaseModel):
    r"""
    ie-HGCN from paper `Interpretable and Efficient Heterogeneous Graph Convolutional Network
    <https://arxiv.org/pdf/2005.13183.pdf>`__.

    `Source Code Link <https://github.com/kepsail/ie-HGCN>`_
    
    The core part of ie-HGCN, the calculating flow of projection, object-level aggregation and type-level aggregation in
    a specific type block.

    Projection
    
    .. math::
        Y^{Self-\Omega }=H^{\Omega} \cdot W^{Self-\Omega} \quad (1)-1

        Y^{\Gamma - \Omega}=H^{\Gamma} \cdot W^{\Gamma - \Omega} , \Gamma \in N_{\Omega} \quad (1)-2

    Object-level Aggregation
    
    .. math::
        Z^{ Self - \Omega } = Y^{ Self - \Omega}=H^{\Omega} \cdot W^{Self - \Omega} \quad (2)-1

        Z^{\Gamma - \Omega}=\hat{A}^{\Omega-\Gamma} \cdot Y^{\Gamma - \Omega} = \hat{A}^{\Omega-\Gamma} \cdot H^{\Gamma} \cdot W^{\Gamma - \Omega} \quad (2)-2

    Type-level Aggregation
    
    .. math::
        Q^{\Omega}=Z^{Self-\Omega} \cdot W_q^{\Omega} \quad (3)-1

        K^{Self-\Omega}=Z^{Self -\Omega} \cdot W_{k}^{\Omega} \quad (3)-2

        K^{\Gamma - \Omega}=Z^{\Gamma - \Omega} \cdot W_{k}^{\Omega}, \quad \Gamma \in N_{\Omega} \quad (3)-3

    .. math::
        e^{Self-\Omega}={ELU} ([K^{ Self-\Omega} \| Q^{\Omega}] \cdot w_{a}^{\Omega}) \quad (4)-1

        e^{\Gamma - \Omega}={ELU} ([K^{\Gamma - \Omega} \| Q^{\Omega}] \cdot w_{a}^{\Omega}), \Gamma \in N_{\Omega} \quad (4)-2

    .. math::
        [a^{Self-\Omega}\|a^{1 - \Omega}\| \ldots . a^{\Gamma - \Omega}\|\ldots\| a^{|N_{\Omega}| - \Omega}] \\
        = {softmax}([e^{Self - \Omega}\|e^{1 - \Omega}\| \ldots\|e^{\Gamma - \Omega}\| \ldots \| e^{|N_{\Omega}| - \Omega}]) \quad (5)

    .. math::
        H_{i,:}^{\Omega \prime}=\sigma(a_{i}^{Self-\Omega} \cdot Z_{i,:}^{Self-\Omega}+\sum_{\Gamma \in N_{\Omega}} a_{i}^{\Gamma - \Omega} \cdot Z_{i,:}^{\Gamma - \Omega}) \quad (6)
    
    Parameters
    ----------
    num_layers: int
        the number of layers
    in_dim: int
        the input dimension
    hidden_dim: int
        the hidden dimension
    out_dim: int
        the output dimension
    attn_dim: int
        the dimension of attention vector
    ntypes: list
        the node type of a heterogeneous graph
    etypes: list
        the edge type of a heterogeneous graph
    bias: boolean
        whether we need bias vector
    batchnorm: boolean
        whether we need batchnorm
    dropout: float
        the drop out rate
    """
    @classmethod
    def build_model_from_args(cls, args, hg:dgl.DGLGraph):
        return cls(args.num_layers,
                   args.hidden_dim,
                   args.out_dim,
                   args.attn_dim,
                   hg.ntypes,
                   hg.etypes,
                   args.bias,
                   args.batchnorm,
                   args.dropout
                   )

    def __init__(self, num_layers, hidden_dim, out_dim, attn_dim, ntypes, etypes, bias, batchnorm, dropout):
        super(ieHGCN, self).__init__()
        self.num_layers = num_layers
        self.activation = F.elu
        self.hgcn_layers = nn.ModuleList()

        for i in range(0, num_layers - 1):
            self.hgcn_layers.append(
                ieHGCNConv(
                    hidden_dim,
                    hidden_dim,
                    attn_dim,
                    ntypes,
                    etypes,
                    self.activation,
                    bias,
                    batchnorm,
                    dropout
                )
            )
        
        self.hgcn_layers.append(
            ieHGCNConv(
                hidden_dim,
                out_dim,
                attn_dim,
                ntypes,
                etypes,
                None,
                False,
                False,
                0.0
            )
        )

    def forward(self, hg, h_dict):
        """
        The forward part of the ieHGCN.
        
        Parameters
        ----------
        hg : object
            the dgl heterogeneous graph
        h_dict: dict
            the feature dict of different node types
            
        Returns
        -------
        dict
            The embeddings after the output projection.
        """
        if hasattr(hg, "ntypes"):
            for l in range(self.num_layers):
                h_dict = self.hgcn_layers[l](hg, h_dict)
        else:
            for layer, block in zip(self.hgcn_layers, hg):
                h_dict = layer(block, h_dict)
        return h_dict

class ieHGCNConv(nn.Module):
    r"""
    The ieHGCN convolution layer.

    Parameters
    ----------
    in_size: int
        the input dimension
    out_size: int
        the output dimension
    attn_size: int
        the dimension of attention vector
    ntypes: list
        the node type list of a heterogeneous graph
    etypes: list
        the edge type list of a heterogeneous graph
    activation: str
        the activation function
    bias: boolean
        whether we need bias vector
    batchnorm: boolean
        whether we need batchnorm
    dropout: float
        the drop out rate
    """
    def __init__(self, in_size, out_size, attn_size, ntypes, etypes, activation = F.elu, 
                 bias = False, batchnorm = False, dropout = 0.0):
        super(ieHGCNConv, self).__init__()
        self.bias = bias
        self.batchnorm = batchnorm
        self.dropout = dropout
        node_size = {}
        for ntype in ntypes:
            node_size[ntype] = in_size
        attn_vector = {}
        for ntype in ntypes:
            attn_vector[ntype] =  attn_size
        self.W_self = dglnn.HeteroLinear(node_size, out_size)
        self.W_al = dglnn.HeteroLinear(attn_vector, 1)
        self.W_ar = dglnn.HeteroLinear(attn_vector, 1)
        
        # self.conv = dglnn.HeteroGraphConv({
        #     etype: dglnn.GraphConv(in_size, out_size, norm = 'right', weight = True, bias = True)
        #     for etype in etypes
        # })
        self.in_size = in_size
        self.out_size = out_size
        self.attn_size = attn_size
        mods = {
            etype: dglnn.GraphConv(in_size, out_size, norm = 'right', 
                                   weight = True, bias = True, allow_zero_in_degree = True)
            for etype in etypes
            }
        self.mods = nn.ModuleDict(mods)
        
        self.linear_q = nn.ModuleDict({ntype: nn.Linear(out_size, attn_size) for ntype in ntypes})
        self.linear_k = nn.ModuleDict({ntype: nn.Linear(out_size, attn_size) for ntype in ntypes})
        
        self.activation = activation
        if batchnorm:
            self.bn = nn.BatchNorm1d(out_size)
        if bias:
            self.h_bias = nn.Parameter(torch.Tensor(out_size))
            nn.init.zeros_(self.h_bias)      
        self.dropout = nn.Dropout(dropout)

    def forward(self, hg, h_dict):
        """
        The forward part of the ieHGCNConv.
        
        Parameters
        ----------
        hg : object or list[block]
            the dgl heterogeneous graph or the list of blocks
        h_dict: dict
            the feature dict of different node types
            
        Returns
        -------
        dict
            The embeddings after final aggregation.
        """
        outputs = {ntype: [] for ntype in hg.dsttypes}
        if hg.is_block:
            src_inputs = h_dict
            dst_inputs = {k: v[:hg.number_of_dst_nodes(k)] for k, v in h_dict.items()}
        else:
            src_inputs = h_dict
            dst_inputs = h_dict
        with hg.local_scope():
            hg.ndata['h'] = h_dict
            # formulas (2)-1
            dst_inputs = self.W_self(dst_inputs)
            query = {}
            key = {}
            attn = {}
            attention = {}
            
            # formulas (3)-1 and (3)-2
            for ntype in hg.dsttypes:
                query[ntype] = self.linear_q[ntype](dst_inputs[ntype])
                key[ntype] = self.linear_k[ntype](dst_inputs[ntype])
            # formulas (4)-1
            h_l = self.W_al(key)
            h_r = self.W_ar(query)
            for ntype in hg.dsttypes:
                attention[ntype] = F.elu(h_l[ntype] + h_r[ntype])
                attention[ntype] = attention[ntype].unsqueeze(0)
            
            for srctype, etype, dsttype in hg.canonical_etypes:
                rel_graph = hg[srctype, etype, dsttype]
                if srctype not in h_dict:
                    continue
                # formulas (2)-2
                dstdata = self.mods[etype](
                    rel_graph,
                    (src_inputs[srctype], dst_inputs[dsttype])
                )
                outputs[dsttype].append(dstdata)
                # formulas (3)-3
                attn[dsttype] = self.linear_k[dsttype](dstdata)
                # formulas (4)-2
                h_attn = self.W_al(attn)
                attn.clear()
                edge_attention = F.elu(h_attn[dsttype] + h_r[dsttype])
                attention[dsttype] = torch.cat((attention[dsttype], edge_attention.unsqueeze(0)))

            # formulas (5)
            for ntype in hg.dsttypes:
                attention[ntype] = F.softmax(attention[ntype], dim = 0)

            # formulas (6)
            rst = {ntype: 0 for ntype in hg.dsttypes}
            for ntype, data in outputs.items():
                data = [dst_inputs[ntype]] + data
                if len(data) != 0:
                    for i in range(len(data)):
                        aggregation = torch.mul(data[i], attention[ntype][i])
                        rst[ntype] = aggregation + rst[ntype]
                
            # h = self.conv(hg, hg.ndata['h'], aggregate = self.my_agg_func)
        def _apply(ntype, h):
            if self.bias:
                h = h + self.h_bias
            if self.activation:
                h = self.activation(h)
            if self.batchnorm:
                h = self.bn(h)
            return self.dropout(h)
            
        return {ntype: _apply(ntype, feat) for ntype, feat in rst.items()}


import dgl
import torch as th
import torch.nn as nn
import torch.nn.functional as F
from dgl.nn.pytorch import GraphConv, EdgeWeightNorm
from ..utils import transform_relation_graph_list
from . import BaseModel, register_model


@register_model('GTN')
class GTN(BaseModel):
    r"""
        GTN from paper `Graph Transformer Networks <https://arxiv.org/abs/1911.06455>`__
        in NeurIPS_2019. You can also see the extension paper `Graph Transformer
        Networks: Learning Meta-path Graphs to Improve GNNs <https://arxiv.org/abs/2106.06218.pdf>`__.

        `Code from author <https://github.com/seongjunyun/Graph_Transformer_Networks>`__.

        Given a heterogeneous graph :math:`G` and its edge relation type set :math:`\mathcal{R}`.Then we extract
        the single relation adjacency matrix list. In that, we can generate combination adjacency matrix by conv
        the single relation adjacency matrix list. We can generate :math:'l-length' meta-path adjacency matrix
        by multiplying combination adjacency matrix. Then we can generate node representation using a GCN layer.

        Parameters
        ----------
        num_edge_type : int
            Number of relations.
        num_channels : int
            Number of conv channels.
        in_dim : int
            The dimension of input feature.
        hidden_dim : int
            The dimension of hidden layer.
        num_class : int
            Number of classification type.
        num_layers : int
            Length of hybrid metapath.
        category : string
            Type of predicted nodes.
        norm : bool
            If True, the adjacency matrix will be normalized.
        identity : bool
            If True, the identity matrix will be added to relation matrix set.

    """
    @classmethod
    def build_model_from_args(cls, args, hg):
        if args.identity:
            num_edge_type = len(hg.canonical_etypes) + 1
        else:
            num_edge_type = len(hg.canonical_etypes)
        # add self-loop edge
        return cls(num_edge_type=num_edge_type, num_channels=args.num_channels,
                   in_dim=args.hidden_dim, hidden_dim=args.hidden_dim, num_class=args.out_dim,
                   num_layers=args.num_layers, category=args.category, norm=args.norm_emd_flag, identity=args.identity)

    def __init__(self, num_edge_type, num_channels, in_dim, hidden_dim, num_class, num_layers, category, norm,
                 identity):
        super(GTN, self).__init__()
        self.num_edge_type = num_edge_type
        self.num_channels = num_channels
        self.in_dim = in_dim
        self.hidden_dim = hidden_dim
        self.num_class = num_class
        self.num_layers = num_layers
        self.is_norm = norm
        self.category = category
        self.identity = identity

        layers = []
        for i in range(num_layers):
            if i == 0:
                layers.append(GTLayer(num_edge_type, num_channels, first=True))
            else:
                layers.append(GTLayer(num_edge_type, num_channels, first=False))
        self.layers = nn.ModuleList(layers)
        self.gcn = GraphConv(in_feats=self.in_dim, out_feats=hidden_dim, norm='none', activation=F.relu)
        self.norm = EdgeWeightNorm(norm='right')
        self.linear1 = nn.Linear(self.hidden_dim * self.num_channels, self.hidden_dim)
        self.linear2 = nn.Linear(self.hidden_dim, self.num_class)
        self.category_idx = None
        self.A = None
        self.h = None

    def normalization(self, H):
        norm_H = []
        for i in range(self.num_channels):
            g = H[i]
            g = dgl.remove_self_loop(g)
            g.edata['w_sum'] = self.norm(g, g.edata['w_sum'])
            norm_H.append(g)
        return norm_H

    def forward(self, hg, h):
        with hg.local_scope():
            hg.ndata['h'] = h
            # * =============== Extract edges in original graph ================
            if self.category_idx is None:
                self.A, h, self.category_idx = transform_relation_graph_list(hg, category=self.category,
                                                                             identity=self.identity)
            else:
                g = dgl.to_homogeneous(hg, ndata='h')
                h = g.ndata['h']
            # X_ = self.gcn(g, self.h)
            A = self.A
            # * =============== Get new graph structure ================
            for i in range(self.num_layers):
                if i == 0:
                    H, W = self.layers[i](A)
                else:
                    H, W = self.layers[i](A, H)
                if self.is_norm == True:
                    H = self.normalization(H)
                # Ws.append(W)
            # * =============== GCN Encoder ================
            for i in range(self.num_channels):
                g = dgl.remove_self_loop(H[i])
                edge_weight = g.edata['w_sum']
                g = dgl.add_self_loop(g)
                edge_weight = th.cat((edge_weight, th.full((g.number_of_nodes(),), 1, device=g.device)))
                edge_weight = self.norm(g, edge_weight)
                if i == 0:
                    X_ = self.gcn(g, h, edge_weight=edge_weight)
                else:
                    X_ = th.cat((X_, self.gcn(g, h, edge_weight=edge_weight)), dim=1)
            X_ = self.linear1(X_)
            X_ = F.relu(X_)
            y = self.linear2(X_)
            return {self.category: y[self.category_idx]}


class GTLayer(nn.Module):
    r"""
        CTLayer multiply each combination adjacency matrix :math:`l` times to a :math:`l-length`
        meta-paths adjacency matrix.

        The method to generate :math:`l-length` meta-path adjacency matrix can be described as:

        .. math::
            A_{(l)}=\Pi_{i=1}^{l} A_{i}

        where :math:`A_{i}` is the combination adjacency matrix generated by GT conv.

        Parameters
        ----------
            in_channels: int
                The input dimension of GTConv which is numerically equal to the number of relations.
            out_channels: int
                The input dimension of GTConv which is numerically equal to the number of channel in GTN.
            first: bool
                If true, the first combination adjacency matrix multiply the combination adjacency matrix.

    """
    def __init__(self, in_channels, out_channels, first=True):
        super(GTLayer, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.first = first
        if self.first:
            self.conv1 = GTConv(in_channels, out_channels)
            self.conv2 = GTConv(in_channels, out_channels)
        else:
            self.conv1 = GTConv(in_channels, out_channels)

    def forward(self, A, H_=None):
        if self.first:
            result_A = self.conv1(A)
            result_B = self.conv2(A)
            W = [(F.softmax(self.conv1.weight, dim=1)).detach(), (F.softmax(self.conv2.weight, dim=1)).detach()]
        else:
            result_A = H_
            result_B = self.conv1(A)
            W = [(F.softmax(self.conv1.weight, dim=1)).detach()]
        H = []
        for i in range(len(result_A)):
            g = dgl.adj_product_graph(result_A[i], result_B[i], 'w_sum')
            H.append(g)
        return H, W


class GTConv(nn.Module):
    r"""
        We conv each sub adjacency matrix :math:`A_{R_{i}}` to a combination adjacency matrix :math:`A_{1}`:

        .. math::
            A_{1} = conv\left(A ; W_{c}\right)=\sum_{R_{i} \in R} w_{R_{i}} A_{R_{i}}

        where :math:`R_i \subseteq \mathcal{R}` and :math:`W_{c}` is the weight of each relation matrix
    """

    def __init__(self, in_channels, out_channels, softmax_flag=True):
        super(GTConv, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.weight = nn.Parameter(th.Tensor(out_channels, in_channels))
        self.softmax_flag = softmax_flag
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.normal_(self.weight, std=0.01)

    def forward(self, A):
        if self.softmax_flag:
            Filter = F.softmax(self.weight, dim=1)
        else:
            Filter = self.weight
        num_channels = Filter.shape[0]
        results = []
        for i in range(num_channels):
            for j, g in enumerate(A):
                A[j].edata['w_sum'] = g.edata['w'] * Filter[i][j]
            sum_g = dgl.adj_sum_graph(A, 'w_sum')
            results.append(sum_g)
        return results



from dataset import build_dataset
from dataset import AcademicDataset
from dataset import Mg2vecDataSet
import dgl
from dgl.data import DGLDataset
from dgl import transforms as T
from dataset import AsLinkPredictionDataset, AsNodeClassificationDataset

def construct_dataset(name,task):
    if task == 'node_classification':
        if name in [
                    'acm4NSHE', 'acm4GTN', 'academic4HetGNN', 'acm4HetGNN','acm_han_raw',
                    'acm4HeCo', 'dblp4MAGNN', 'imdb4MAGNN', 'imdb4GTN',
                    'acm4NARS', 'yelp4HeGAN',
                    'HGBn-ACM', 'HGBn-DBLP', 'HGBn-Freebase', 'HGBn-IMDB',
                    'alircd_session1',
                    'ohgbn-Freebase', 'ohgbn-yelp2', 'ohgbn-acm', 'ohgbn-imdb',
                    'dblp4GTN',
                    'HNE-PubMed', #Fixed: should be in 'node_classification' task, instead of 'link_prediction'
                    'ogbn-mag',# Download speed low ...
                    'aifb', 'mutag', 'bgs', 'am',
                    'alircd_small',
                    'ICDM',
                    ]:
            if name == "acm4HetGNN":
                name = 'academic4HetGNN'
            return build_dataset(dataset=name, task=task, logger=None)
        
        elif name  == "academic4HetGNN":
            return AcademicDataset("academic4HetGNN")
        
        # Fixed: change 'dblp4Mg2vec_4'、'dblp4Mg2vec_5' into 'dblp4mg2vec_4'、'dblp4mg2vec_5'.
        # No model named 'mg2vec', corresponding dataset had no node labels.

    elif task == 'link_prediction':
        #amazon4SLICE can't use for now.
        if name in ['amazon4SLICE', 
                    'MTWM', 'HGBl-ACM',
                    'HGBl-DBLP', 'HGBl-IMDB',
                    'wn18', 'FB15k', 'FB15k-237',
                    'HGBl-amazon', 'HGBl-LastFM', 'HGBl-PubMed',
                    'ohgbl-MTWM', 'ohgbl-yelp1', 'ohgbl-yelp2', 'ohgbl-Freebase']:   
            return build_dataset(dataset=name, task=task, logger=None)
        elif name in ['DoubanMovie']: # Fixed, no class registered in hin_link_prediction
            ds = AcademicDataset('DoubanMovie')
            ds.g = ds[0]
            return ds
        

    elif task == 'recommendation': 
        if name in ['LastFM4KGCN','yelp4rec']:
            return build_dataset(dataset=name, task=task, logger=None)

    elif task == 'edge_classification':
        if name in ['dblp4Mg2vec_4', 'dblp4Mg2vec_5']:
            ds = Mg2vecDataSet(name=name)
            ds.g = ds[0]
            return ds
class MyDataset(DGLDataset):
    def __init__(self,name,path,reverse=True):
        self.path = path
        self.reverse = reverse
        super().__init__(name=name)
        

    def process(self):
        gs, _ = dgl.load_graphs(self.path)
        if(self.reverse):
            self._g = T.AddReverse()(gs[0])
        else:
            self._g = gs[0]

    def __getitem__(self, idx):
        return self._g

    def __len__(self):
        return 1




import torch.nn as nn
from . import BaseTask, register_task
from ..dataset import build_dataset
from ..utils import Evaluator
from sklearn.metrics import roc_auc_score, average_precision_score

@register_task("abnorm_event_detection")
class AbnormalEventDetection(BaseTask):
    """Abnormal event detection task"""
    def __init__(self, args):
        super(AbnormalEventDetection, self).__init__()
        self.logger = args.logger
        self.n_dataset = args.dataset
        self.dataset = build_dataset(args.dataset+'4'+args.model, 'abnorm_event_detection', logger=self.logger)
        self.logger = args.logger
        self.args = args

    def get_graph(self):
        return self.dataset.g

    def get_loss_fn(self):
        return None

    def get_evaluator(self, name):
        if name == 'AP':
            return average_precision_score
        elif name == 'AUC':
            return roc_auc_score
        else:
            ValueError('evaluator name must be in AP and AUC')

    def evaluate(self, *args, **kwargs):
        pass


import numpy as np
from collections import Counter
import torch as th

__all__ = ['plot_degree_dist', 'plot_portion', 'plot_number_metapath']


def plot_portion(g, save_path=None, **kwargs):
    import matplotlib.pyplot as plt
    ntypes = g.ntypes
    num_nodes = []
    for ntype in ntypes:
        print(ntype)
        num_nodes.append(g.num_nodes(ntype))
        plt.style.use('ggplot')
    plt.pie(num_nodes, radius=1, autopct='%.3f%%', labels=ntypes, wedgeprops=dict(width=0.4))
    if save_path is None:
        plt.show()
    else:
        plt.savefig(save_path)


def plot_degree_dist(g, save_path=None, **kwargs):
    import matplotlib.pyplot as plt

    canonical_etypes = g.canonical_etypes
    ntypes = g.ntypes
    x_list = []
    y_list = []
    xscales = []
    yscales = []
    for ntype in ntypes:
        print(ntype)
        degrees = np.array([0] * g.num_nodes(ntype))
        for canonical_etype in canonical_etypes:
            etype = canonical_etype[1]
            if ntype in canonical_etype[0]:  # 出度
                out_degree = g.out_degrees(g.nodes(ntype), etype=etype)
                out_degree = out_degree.numpy()
                degrees += out_degree
            if ntype in canonical_etype[2]:  # 入度
                in_degree = g.in_degrees(g.nodes(ntype), etype=etype)
                in_degree = in_degree.numpy()
                degrees += in_degree
        degree_counts = Counter(degrees)
        x, y = zip(*degree_counts.items())
        xscales.append(max(x))
        yscales.append(max(y))
        x_list.append(x)
        y_list.append(y)

    plt.figure(1)
    plt.style.use('ggplot')
    # prep axes
    plt.xlabel('Degree')
    plt.xscale('log')
    plt.xlim(1, max(xscales))

    plt.ylabel('Number_of_Nodes')
    plt.yscale('log')
    plt.ylim(1, max(yscales))
    # do plot
    for i in range(len(ntypes)):
        plt.scatter(x_list[i], y_list[i], marker='*')
    plt.legend(ntypes)
    if save_path is None:
        plt.show()
    else:
        plt.savefig(save_path)


def plot_number_metapath(g, meta_paths_dict, save_path=None, **kwargs):
    import matplotlib.pyplot as plt

    meta_path_names = []
    meta_path_nums = []
    for meta_path_name, meta_path in meta_paths_dict.items():
        meta_path_names.append(meta_path_name)
        for i, etype in enumerate(meta_path):
            if(isinstance(etype,list)):
                _new_type = (etype[0],etype[1],etype[2])
                etype = _new_type
            if i == 0:
                adj = g.adj(etype=etype)
            else:
                adj = th.sparse.mm(adj, g.adj(etype=etype))
        meta_path_nums.append(int(th.sparse.sum(adj)))

    plt.figure(1)
    plt.style.use('ggplot')
    # prep axes
    plt.xlabel('Meta-path')
    plt.ylabel('Number_of_Meta-path')
    plt.bar(meta_path_names, meta_path_nums,
            color=['gold', 'yellowgreen', 'lightseagreen', 'cornflowerblue', 'royalblue'],
            width=0.5)
    if save_path is None:
        plt.show()
    else:
        plt.savefig(save_path)


def vis_emb(emb, label, save_path=None):
    """
    This is a demo of embedding visualization.
    Arguments:
        `emb` (numpy.ndarray): node embeddings 
        `label` (numpy.ndarray): node labels
        `save_path` (str): figure path to save. 
    
    After training is done, using like following in your trainer:
    ``` python
    from openhgnn.utils import vis_emb
    emb = self.model.get_emb()
    label = self.task.get_labels()
    vis_emb(emb, label, "path_to_save")
    ```
    """

    import matplotlib.pyplot as plt
    from sklearn import manifold

    tsne = manifold.TSNE(n_components=2, init='pca', random_state=501)
    X_tsne = tsne.fit_transform(emb)
    x_min, x_max = X_tsne.min(0), X_tsne.max(0)
    X_norm = (X_tsne - x_min) / (x_max - x_min)
    color_values = label/len(set(label))
    plt.scatter(X_norm[:,0], X_norm[:,1], cmap=plt.get_cmap('jet'), c=color_values)

    if save_path is None:
        plt.show()
    else:
        plt.savefig(save_path)



import torch.nn.functional as F
import torch.nn as nn
from . import BaseTask, register_task
from ..dataset import build_dataset
from ..utils import Evaluator


@register_task("edge_classification")
class EdgeClassification(BaseTask):
    r"""
    Edge classification tasks.

    Attributes
    -----------
    dataset : EdgeClassificationDataset
        Task-related dataset

    evaluator : Evaluator
        offer evaluation metric


    Methods
    ---------
    get_graph :
        return a graph
    get_loss_fn :
        return a loss function
    """

    def __init__(self, args):
        super(EdgeClassification, self).__init__()
        self.logger = args.logger
        self.dataset = build_dataset(args.dataset, 'edge_classification', logger=self.logger)
        # self.evaluator = Evaluator()
        self.logger = args.logger
        if hasattr(args, 'validation'):
            self.train_idx, self.val_idx, self.test_idx = self.dataset.get_split(args.validation)
        else:
            self.train_idx, self.val_idx, self.test_idx = self.dataset.get_split(validation=False)
        self.evaluator = Evaluator(args.seed)
        self.labels = self.dataset.get_labels()
        self.multi_label = self.dataset.multi_label

        if hasattr(args, 'evaluation_metric'):
            self.evaluation_metric = args.evaluation_metric
        else:
            if args.dataset in ['dblp4Mg2vec']:
                self.evaluation_metric = 'acc'
            else:
                self.evaluation_metric = 'f1'

    def get_graph(self):
        return self.dataset.g

    def get_loss_fn(self):
        if self.multi_label:
            return nn.BCEWithLogitsLoss()
        return F.cross_entropy

    def get_evaluator(self, name):
        if name == 'acc':
            return self.evaluator.cal_acc
        elif name == 'f1_lr':
            return self.evaluator.nc_with_LR
        elif name == 'f1':
            return self.evaluator.f1_node_classification

    def evaluate(self, logits, mode='test', info=True):
        if mode == 'test':
            mask = self.test_idx
        elif mode == 'valid':
            mask = self.val_idx
        elif mode == 'train':
            mask = self.train_idx

        if self.multi_label:
            pred = (logits[mask].cpu().numpy() > 0).astype(int)
        else:
            pred = logits[mask].argmax(dim=1).to('cpu')

        if self.evaluation_metric == 'acc':
            acc = self.evaluator.cal_acc(self.labels[mask], pred)
            return dict(Accuracy=acc)
        elif self.evaluation_metric == 'f1':
            f1_dict = self.evaluator.f1_node_classification(self.labels[mask], pred)
            return f1_dict
        else:
            raise ValueError('The evaluation metric is not supported!')

    def downstream_evaluate(self, logits, evaluation_metric):
        if evaluation_metric == 'acc_f1':
            micro_f1, macro_f1, acc = self.evaluator.ec_with_SVC(C=100, gamma='auto', emd=logits, labels=self.labels,
                                                                 train_idx=self.train_idx, test_idx=self.test_idx)
            return dict(Macro_f1=macro_f1, Mirco_f1=micro_f1, Accuracy=acc)

    def get_split(self):
        return self.train_idx, self.val_idx, self.test_idx

    def get_labels(self):
        return self.labels



from pathlib import Path
import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset

FOLD = Path(__file__).resolve().parent / "meirec"


class MeiRECDataset(Dataset):

    def __init__(self, phase):
        assert phase in ['train', 'test']
        if phase == 'train':
            self.data_path = FOLD / "train_data.txt"
        else:
            self.data_path = FOLD / "test_data.txt"
        self.load_data(self.data_path)

    def load_data(self, path):
        self.features_list = []
        self.label_list = []
        with open(path, 'r', encoding='utf-8') as f:
            for line in f.readlines():
                feature, label = line.split()
                self.features_list.append(feature)
                self.label_list.append(float(label))
        self.data_len = len(self.features_list)
        # print(self.data_len)

    def __getitem__(self, item):
        index = item
        feature_list = []
        feature = self.features_list[index]
        features_split = feature.split(',')

        for item in features_split[:81]:
            feature_list.append(float(item))
        for item in features_split[81:]:
            feature_list.append(int(item))

        return {
            'data': torch.from_numpy(np.array(feature_list)),
            'labels': self.label_list[index]
        }

    def __len__(self):
        return self.data_len


def get_data_loader(type, batch_size=64, num_workers=0):
    dataset = MeiRECDataset(type)
    dataloader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        drop_last=True,
    )
    return dataloader




import torch
import torch.nn as nn
import torch.nn.functional as F
from openhgnn.models import BaseModel, register_model
import numpy as np

@register_model('DiffMG')
class DiffMG(BaseModel):

    @classmethod

    def build_model_from_args(cls, args,hg):
        args.search_model = search_model
        return cls

    def __init__(self, in_dims, n_hid, n_steps, dropout=None, attn_dim=64, use_norm=True, out_nl=True):
        super(DiffMG, self).__init__()
        self.n_hid = n_hid
        self.ws = nn.ModuleList()
        assert (isinstance(in_dims, list))
        for i in range(len(in_dims)):
            self.ws.append(nn.Linear(in_dims[i], n_hid))
        assert (isinstance(n_steps, list))
        self.metas = nn.ModuleList()
        for i in range(len(n_steps)):
            self.metas.append(Cell(n_steps[i], n_hid, n_hid, use_norm=use_norm, use_nl=out_nl))

        # * [Optional] Combine more than one meta graph?
        self.attn_fc1 = nn.Linear(n_hid, attn_dim)
        self.attn_fc2 = nn.Linear(attn_dim, 1)

        self.feats_drop = nn.Dropout(dropout) if dropout is not None else lambda x: x

    def forward(self, node_feats, node_types, adjs, idxes_seq, idxes_res, gpu):
        if gpu > -1:
            hid = torch.zeros((len(node_types), self.n_hid)).cuda()
        else:
            hid = torch.zeros((len(node_types), self.n_hid))
        for i in range(len(node_feats)):
            hid[node_types == i] = self.ws[i](node_feats[i])
        hid = self.feats_drop(hid)
        temps = [];
        attns = []
        for i, meta in enumerate(self.metas):
            hidi = meta(hid, adjs, idxes_seq[i], idxes_res[i])
            temps.append(hidi)
            attni = self.attn_fc2(torch.tanh(self.attn_fc1(temps[-1])))
            attns.append(attni)

        hids = torch.stack(temps, dim=0).transpose(0, 1)
        attns = F.softmax(torch.cat(attns, dim=-1), dim=-1)
        out = (attns.unsqueeze(dim=-1) * hids).sum(dim=1)
        return out
class Op(nn.Module):
    def __init__(self):
        super(Op, self).__init__()

    # 定义图卷积的操作，其中adjs是邻接矩阵的列表，idx是当前需要使用的邻接矩阵的索引。
    def forward(self, x, adjs, idx):
        return torch.spmm(adjs[idx], x)


class Cell(nn.Module):

    def __init__(self, n_step, n_hid_prev, n_hid, use_norm=True, use_nl=True):
        super(Cell, self).__init__()

        self.affine = nn.Linear(n_hid_prev, n_hid)
        self.n_step = n_step
        self.norm = nn.LayerNorm(n_hid) if use_norm is True else lambda x: x
        self.use_nl = use_nl
        self.ops_seq = nn.ModuleList()
        self.ops_res = nn.ModuleList()
        for i in range(self.n_step):
            self.ops_seq.append(Op())
        for i in range(1, self.n_step):
            for j in range(i):
                self.ops_res.append(Op())

    def forward(self, x, adjs, idxes_seq, idxes_res):

        x = self.affine(x)
        states = [x]
        offset = 0
        for i in range(self.n_step):
            seqi = self.ops_seq[i](states[i], adjs[:-1], idxes_seq[i])  # ! exclude zero Op
            resi = sum(self.ops_res[offset + j](h, adjs, idxes_res[offset + j]) for j, h in enumerate(states[:i]))
            offset += i
            states.append(seqi + resi)
        # assert(offset == len(self.ops_res))

        output = self.norm(states[-1])
        if self.use_nl:
            output = F.gelu(output)
        return output

class Op1(nn.Module):
    '''
    operation for one link in the DAG search space
    '''

    def __init__(self):
        super(Op1, self).__init__()

    def forward(self, x, adjs, ws, idx):
        # assert(ws.size(0) == len(adjs))
        return ws[idx] * torch.spmm(adjs[idx], x)


class Cell1(nn.Module):
    '''
    the DAG search space
    '''

    def __init__(self, n_step, n_hid_prev, n_hid, cstr, use_norm=True, use_nl=True):
        super(Cell1, self).__init__()

        self.affine = nn.Linear(n_hid_prev, n_hid)
        self.n_step = n_step  # * number of intermediate states (i.e., K)
        self.norm = nn.LayerNorm(n_hid, elementwise_affine=False) if use_norm is True else lambda x: x
        self.use_nl = use_nl
        assert (isinstance(cstr, list))
        self.cstr = cstr  # * type constraint

        self.ops_seq = nn.ModuleList()  # * state (i - 1) -> state i, 1 <= i < K
        for i in range(1, self.n_step):
            self.ops_seq.append(Op1())
        self.ops_res = nn.ModuleList()  # * state j -> state i, 0 <= j < i - 1, 2 <= i < K
        for i in range(2, self.n_step):
            for j in range(i - 1):
                self.ops_res.append(Op1())

        self.last_seq = Op1()  # * state (K - 1) -> state K
        self.last_res = nn.ModuleList()  # * state i -> state K, 0 <= i < K - 1
        for i in range(self.n_step - 1):
            self.last_res.append(Op1())

    def forward(self, x, adjs, ws_seq, idxes_seq, ws_res, idxes_res):
        # assert(isinstance(ws_seq, list))
        # assert(len(ws_seq) == 2)
        x = self.affine(x)
        states = [x]
        offset = 0
        for i in range(self.n_step - 1):
            seqi = self.ops_seq[i](states[i], adjs[:-1], ws_seq[0][i], idxes_seq[0][i])  # ! exclude zero Op
            resi = sum(self.ops_res[offset + j](h, adjs, ws_res[0][offset + j], idxes_res[0][offset + j]) for j, h in
                       enumerate(states[:i]))
            offset += i
            states.append(seqi + resi)
        # assert(offset == len(self.ops_res))

        adjs_cstr = [adjs[i] for i in self.cstr]
        out_seq = self.last_seq(states[-1], adjs_cstr, ws_seq[1], idxes_seq[1])
        adjs_cstr.append(adjs[-1])
        out_res = sum(self.last_res[i](h, adjs_cstr, ws_res[1][i], idxes_res[1][i]) for i, h in enumerate(states[:-1]))
        output = self.norm(out_seq + out_res)
        if self.use_nl:
            output = F.gelu(output)
        return output


class search_model(nn.Module):

    def __init__(self, in_dims, n_hid, n_adjs, n_steps, cstr, attn_dim=64, use_norm=True, out_nl=True):
        super(search_model, self).__init__()
        self.cstr = cstr
        self.n_adjs = n_adjs
        self.n_hid = n_hid
        self.ws = nn.ModuleList()  # * node type-specific transformation
        assert (isinstance(in_dims, list))
        for i in range(len(in_dims)):
            self.ws.append(nn.Linear(in_dims[i], n_hid))
        assert (isinstance(n_steps, list))  # * [optional] combine more than one meta graph?
        self.metas = nn.ModuleList()
        for i in range(len(n_steps)):
            self.metas.append(Cell1(n_steps[i], n_hid, n_hid, cstr, use_norm=use_norm, use_nl=out_nl))

        self.as_seq = []  # * arch parameters for ops_seq
        self.as_last_seq = []  # * arch parameters for last_seq
        for i in range(len(n_steps)):
            if n_steps[i] > 1:
                ai = 1e-3 * torch.randn(n_steps[i] - 1, n_adjs - 1)  # ! exclude zero Op
                ai = ai
                ai.requires_grad_(True)
                self.as_seq.append(ai)
            else:
                self.as_seq.append(None)
            ai_last = 1e-3 * torch.randn(len(cstr))
            ai_last = ai_last
            ai_last.requires_grad_(True)
            self.as_last_seq.append(ai_last)

        ks = [sum(1 for i in range(2, n_steps[k]) for j in range(i - 1)) for k in range(len(n_steps))]
        self.as_res = []  # * arch parameters for ops_res
        self.as_last_res = []  # * arch parameters for last_res
        for i in range(len(n_steps)):
            if ks[i] > 0:
                ai = 1e-3 * torch.randn(ks[i], n_adjs)
                ai = ai
                ai.requires_grad_(True)
                self.as_res.append(ai)
            else:
                self.as_res.append(None)

            if n_steps[i] > 1:
                ai_last = 1e-3 * torch.randn(n_steps[i] - 1, len(cstr) + 1)
                ai_last = ai_last
                ai_last.requires_grad_(True)
                self.as_last_res.append(ai_last)
            else:
                self.as_last_res.append(None)

        assert (ks[0] + n_steps[0] + (0 if self.as_last_res[0] is None else self.as_last_res[0].size(0)) == (
                    1 + n_steps[0]) * n_steps[0] // 2)

        # * [optional] combine more than one meta graph?
        self.attn_fc1 = nn.Linear(n_hid, attn_dim)
        self.attn_fc2 = nn.Linear(attn_dim, 1)

    def alphas(self):
        alphas = []
        for each in self.as_seq:
            if each is not None:
                alphas.append(each)
        for each in self.as_last_seq:
            alphas.append(each)
        for each in self.as_res:
            if each is not None:
                alphas.append(each)
        for each in self.as_last_res:
            if each is not None:
                alphas.append(each)
        return alphas

    def sample(self, eps):
        '''
        to sample one candidate edge type per link
        '''
        idxes_seq = []
        idxes_res = []
        if np.random.uniform() < eps:
            for i in range(len(self.metas)):
                temp = []
                temp.append(None if self.as_seq[i] is None else torch.randint(low=0, high=self.as_seq[i].size(-1),
                                                                              size=self.as_seq[i].size()[:-1]))
                temp.append(torch.randint(low=0, high=self.as_last_seq[i].size(-1), size=(1,)))
                idxes_seq.append(temp)
            for i in range(len(self.metas)):
                temp = []
                temp.append(None if self.as_res[i] is None else torch.randint(low=0, high=self.as_res[i].size(-1),
                                                                              size=self.as_res[i].size()[:-1]))
                temp.append(
                    None if self.as_last_res[i] is None else torch.randint(low=0, high=self.as_last_res[i].size(-1),
                                                                           size=self.as_last_res[i].size()[:-1]))
                idxes_res.append(temp)
        else:
            for i in range(len(self.metas)):
                temp = []
                temp.append(None if self.as_seq[i] is None else torch.argmax(F.softmax(self.as_seq[i], dim=-1), dim=-1))
                temp.append(torch.argmax(F.softmax(self.as_last_seq[i], dim=-1), dim=-1))
                idxes_seq.append(temp)
            for i in range(len(self.metas)):
                temp = []
                temp.append(None if self.as_res[i] is None else torch.argmax(F.softmax(self.as_res[i], dim=-1), dim=-1))
                temp.append(
                    None if self.as_last_res[i] is None else torch.argmax(F.softmax(self.as_last_res[i], dim=-1),
                                                                          dim=-1))
                idxes_res.append(temp)
        return idxes_seq, idxes_res

    def forward(self, node_feats, node_types, adjs, idxes_seq, idxes_res):
        hid = torch.zeros((len(node_types), self.n_hid))
        for i in range(len(node_feats)):
            hid[node_types == i] = self.ws[i](node_feats[i])
        temps = [];
        attns = []
        for i, meta in enumerate(self.metas):
            ws_seq = []
            ws_seq.append(None if self.as_seq[i] is None else F.softmax(self.as_seq[i], dim=-1))
            ws_seq.append(F.softmax(self.as_last_seq[i], dim=-1))
            ws_res = []
            ws_res.append(None if self.as_res[i] is None else F.softmax(self.as_res[i], dim=-1))
            ws_res.append(None if self.as_last_res[i] is None else F.softmax(self.as_last_res[i], dim=-1))
            hidi = meta(hid, adjs, ws_seq, idxes_seq[i], ws_res, idxes_res[i])
            temps.append(hidi)
            attni = self.attn_fc2(torch.tanh(self.attn_fc1(temps[-1])))
            attns.append(attni)

        hids = torch.stack(temps, dim=0).transpose(0, 1)
        attns = F.softmax(torch.cat(attns, dim=-1), dim=-1)
        out = (attns.unsqueeze(dim=-1) * hids).sum(dim=1)
        return out

    def parse(self):
        '''
        to derive a meta graph indicated by arch parameters
        '''
        idxes_seq, idxes_res = self.sample(0.)
        msg_seq = [];
        msg_res = []
        for i in range(len(idxes_seq)):
            map_seq = [self.cstr[idxes_seq[i][1].item()]]
            msg_seq.append(map_seq if idxes_seq[i][0] is None else idxes_seq[i][0].tolist() + map_seq)
            assert (len(msg_seq[i]) == self.metas[i].n_step)

            temp_res = []
            if idxes_res[i][1] is not None:
                for item in idxes_res[i][1].tolist():
                    if item < len(self.cstr):
                        temp_res.append(self.cstr[item])
                    else:
                        assert (item == len(self.cstr))
                        temp_res.append(self.n_adjs - 1)
                if idxes_res[i][0] is not None:
                    temp_res = idxes_res[i][0].tolist() + temp_res
            assert (len(temp_res) == self.metas[i].n_step * (self.metas[i].n_step - 1) // 2)
            msg_res.append(temp_res)

        return msg_seq, msg_res



import numpy as np
import pandas as pd
import os
import pickle
import dgl
from dgl import function as fn
from dgl.nn.functional import edge_softmax
import torch as th
import torch.nn as nn
import torch.nn.functional as F
from dgl.utils import expand_as_pair
from operator import itemgetter
from . import BaseModel, register_model

from sklearn.model_selection import train_test_split
from sklearn.metrics import f1_score
from sklearn.svm import LinearSVC

'''
model
'''


@register_model('MAGNN')
class MAGNN(BaseModel):
    r"""
    This is the main method of model MAGNN

    Parameters
    ----------
    ntypes: list
        the nodes' types of the dataset
    h_feats: int
        hidden dimension
    inter_attn_feats: int
        the dimension of attention vector in inter-metapath aggregation
    num_heads: int
        the number of heads in intra metapath attention
    num_classes: int
        the number of output classes
    num_layers: int
        the number of hidden layers
    metapath_list: list
        the list of metapaths, e.g ['M-D-M', 'M-A-M', ...],
    edge_type_list: list
        the list of edge types, e.g ['M-A', 'A-M', 'M-D', 'D-M'],
    dropout_rate: float
        the dropout rate of feat dropout and attention dropout
    mp_instances : dict
        the metapath instances indices dict. e.g mp_instances['MAM'] stores MAM instances indices.
    encoder_type: str
        the type of encoder, e.g ['RotateE', 'Average', 'Linear']
    activation: callable activation function
        the activation function used in MAGNN.  default: F.elu

    Notes
    -----
    Please make sure that the please make sure that all the metapath is symmetric, e.g ['MDM', 'MAM' ...] are symmetric,
    while ['MAD', 'DAM', ...] are not symmetric.

    please make sure that the edge_type_list meets the following form:
    [edge_type_1, edge_type_1_reverse, edge_type_2, edge_type_2_reverse, ...], like the example above.

    All the activation in MAGNN are the same according to the codes of author.

    """
    @classmethod
    def build_model_from_args(cls, args, hg):
        ntypes = hg.ntypes
        if args.dataset == 'imdb4MAGNN':
            # build model
            metapath_list = ['M-D-M', 'M-A-M', 'D-M-D', 'D-M-A-M-D', 'A-M-A', 'A-M-D-M-A']
            edge_type_list = ['A-M', 'M-A', 'D-M', 'M-D']
            # in_feats: {'n1type': n1_dim, 'n2type', n2_dim, ...}
            in_feats = {'M': 3066, 'D': 2081, 'A': 5257}
            metapath_idx_dict = mp_instance_sampler(hg, metapath_list, 'imdb4MAGNN')

        elif args.dataset == 'dblp4MAGNN':
            # build model
            metapath_list = ['A-P-A', 'A-P-T-P-A', 'A-P-V-P-A']
            edge_type_list = ['A-P', 'P-A', 'P-T', 'T-P', 'P-V', 'V-P']
            # in_feats: {'n1type': n1_dim, 'n2type', n2_dim, ...}
            in_feats = {'A': 334, 'P': 14328, 'T': 7723, 'V': 20}
            metapath_idx_dict = mp_instance_sampler(hg, metapath_list, 'dblp4MAGNN')

        else:
            raise NotImplementedError("MAGNN on dataset {} has not been implemented".format(args.dataset))

        return cls(ntypes=ntypes,
                   h_feats=args.hidden_dim // args.num_heads,
                   inter_attn_feats=args.inter_attn_feats,
                   num_heads=args.num_heads,
                   num_classes=args.out_dim,
                   num_layers=args.num_layers,
                   metapath_list=metapath_list,
                   edge_type_list=edge_type_list,
                   dropout_rate=args.dropout,
                   encoder_type=args.encoder_type,
                   metapath_idx_dict=metapath_idx_dict)

    def __init__(self, ntypes, h_feats, inter_attn_feats, num_heads, num_classes, num_layers,
                 metapath_list, edge_type_list, dropout_rate, metapath_idx_dict, encoder_type='RotateE',
                 activation=F.elu):
        super(MAGNN, self).__init__()

        self.encoder_type = encoder_type
        self.ntypes = ntypes
        self.h_feats = h_feats
        self.inter_attn_feats = inter_attn_feats
        self.num_heads = num_heads
        self.num_classes = num_classes
        self.num_layers = num_layers
        self.metapath_list = metapath_list
        self.edge_type_list = edge_type_list
        self.activation = activation
        self.backup = {}
        self.is_backup = False

        # input projection
        # self.ntypes = in_feats.keys()
        # self.input_projection = nn.ModuleDict()
        # for ntype in self.ntypes:
        #     self.input_projection[ntype] = nn.Linear(in_features=in_feats[ntype], out_features=h_feats * num_heads)

        # for layer in self.input_projection.values():
        #     nn.init.xavier_normal_(layer.weight, gain=1.414)

        # dropout
        self.feat_drop = nn.Dropout(p=dropout_rate)

        # extract ntypes that have corresponding metapath
        # If there're only metapaths like ['M-A-M', 'M-D-M'], 'A' and 'D' have no metapath, so that 'A' and 'D' shouldn't
        # be considered as nodes that need to aggregate information from metapath.
        self.dst_ntypes = set([metapath.split('-')[0] for metapath in metapath_list])

        # hidden layers
        self.layers = nn.ModuleList()
        for i in range(num_layers - 1):
            self.layers.append(
                MAGNN_layer(in_feats=h_feats, inter_attn_feats=inter_attn_feats, out_feats=h_feats, num_heads=num_heads,
                            metapath_list=metapath_list, ntypes=self.ntypes, edge_type_list=edge_type_list,
                            dst_ntypes=self.dst_ntypes, encoder_type=encoder_type, last_layer=False))

        # output layer
        self.layers.append(
            MAGNN_layer(in_feats=h_feats, inter_attn_feats=inter_attn_feats, out_feats=num_classes, num_heads=num_heads,
                        metapath_list=metapath_list, ntypes=self.ntypes, edge_type_list=edge_type_list,
                        dst_ntypes=self.dst_ntypes, encoder_type=encoder_type, last_layer=True))

        self.metapath_idx_dict = metapath_idx_dict

    def mini_reset_params(self, new_metapth_idx_dict):
        '''
        This method is utilized for reset some parameters including metapath_idx_dict, metapath_list, dst_ntypes...
        Other Parameters like weight matrix don't need to be updated.

        '''
        if not self.is_backup:  # the params of the original graph has not been stored
            self.backup['metapath_idx_dict'] = self.metapath_idx_dict
            self.backup['metapath_list'] = self.metapath_list
            self.backup['dst_ntypes'] = self.dst_ntypes
            self.is_backup = True

        self.metapath_idx_dict = new_metapth_idx_dict
        self.metapath_list = list(new_metapth_idx_dict.keys())
        self.dst_ntypes = set([meta[0] for meta in self.metapath_list])

        for layer in self.layers:
            layer.metapath_list = self.metapath_list
            layer.dst_ntypes = self.dst_ntypes

    def restore_params(self):
        assert self.backup, 'The model.backup is empty'
        self.metapath_idx_dict = self.backup['metapath_idx_dict']
        self.metapath_list = self.backup['metapath_list']
        self.dst_ntypes = self.backup['dst_ntypes']

        for layer in self.layers:
            layer.metapath_list = self.metapath_list
            layer.dst_ntypes = self.dst_ntypes

    def forward(self, g, feat_dict=None):
        r"""
        The forward part of MAGNN

        Parameters
        ----------
        g : object
            the dgl heterogeneous graph
        feat_dict : dict
            the feature matrix dict of different node types, e.g {'M':feat_of_M, 'D':feat_of_D, ...}

        Returns
        -------
        dict
            The predicted logit after the output projection. e.g For the predicted node type, such as M(movie),
            dict['M'] contains the probability that each node is classified as each class. For other node types, such as
            D(director), dict['D'] contains the result after the output projection.

        dict
            The embeddings before the output projection. e.g dict['M'] contains embeddings of every node of M type.
        """

        # hidden layer
        for i in range(self.num_layers - 1):
            h, _ = self.layers[i](feat_dict, self.metapath_idx_dict)
            for key in h.keys():
                h[key] = self.activation(h[key])

        # output layer
        h_output, embedding = self.layers[-1](feat_dict, self.metapath_idx_dict)

        # return h_output, embedding
        return h_output


class MAGNN_layer(nn.Module):
    def __init__(self, in_feats, inter_attn_feats, out_feats, num_heads, metapath_list,
                 ntypes, edge_type_list, dst_ntypes, encoder_type='RotateE', last_layer=False):
        super(MAGNN_layer, self).__init__()
        self.in_feats = in_feats
        self.inter_attn_feats = inter_attn_feats
        self.out_feats = out_feats
        self.num_heads = num_heads
        self.metapath_list = metapath_list  # ['M-D-M', 'M-A-M', ...]
        self.ntypes = ntypes  # ['M', 'D', 'A']
        self.edge_type_list = edge_type_list  # ['M-A', 'A-M', ...]
        self.dst_ntypes = dst_ntypes
        self.encoder_type = encoder_type
        self.last_layer = last_layer

        # in_feats_dst_meta = (feature dimension of dst nodes,
        #                      feature dimension of metapath instances after encoding)
        in_feats_dst_meta = tuple((in_feats, in_feats))

        self.intra_attn_layers = nn.ModuleDict()
        for metapath in self.metapath_list:
            self.intra_attn_layers[metapath] = \
                MAGNN_attn_intra(in_feats=in_feats_dst_meta, out_feats=in_feats, num_heads=num_heads)

        # The linear transformation at the beginning of inter metapath aggregation, including all metapath
        # The attention mechanism in inter metapath aggregation
        self.inter_linear = nn.ModuleDict()
        self.inter_attn_vec = nn.ModuleDict()
        for ntype in dst_ntypes:
            self.inter_linear[ntype] = \
                nn.Linear(in_features=in_feats * num_heads, out_features=inter_attn_feats, bias=True)
            self.inter_attn_vec[ntype] = nn.Linear(in_features=inter_attn_feats, out_features=1, bias=False)
            nn.init.xavier_normal_(self.inter_linear[ntype].weight, gain=1.414)
            nn.init.xavier_normal_(self.inter_attn_vec[ntype].weight, gain=1.414)

        # Some initialization related to encoder
        if encoder_type == 'RotateE':
            # r_vec: [r1, r1_inverse, r2, r2_inverse, ...., rn, rn_inverse]
            r_vec_ = nn.Parameter(th.empty(size=(len(edge_type_list) // 2, in_feats * num_heads // 2, 2)))
            nn.init.xavier_normal_(r_vec_.data, gain=1.414)
            self.r_vec = F.normalize(r_vec_, p=2, dim=2)
            self.r_vec = th.stack([self.r_vec, self.r_vec], dim=1)
            self.r_vec[:, 1, :, 1] = -self.r_vec[:, 1, :, 1]
            self.r_vec = self.r_vec.reshape(r_vec_.shape[0] * 2, r_vec_.shape[1], 2)
            self.r_vec_dict = nn.ParameterDict()
            for i, edge_type in zip(range(len(edge_type_list)), edge_type_list):
                self.r_vec_dict[edge_type] = nn.Parameter(self.r_vec[i])

        # The dimension here does not change because multi-heads conversion has been done before
        # This part is a little different from the original author's codes.
        elif encoder_type == 'Linear':
            self.encoder_linear = \
                nn.Linear(in_features=in_feats * num_heads, out_features=in_feats * num_heads)

        # output layer
        if last_layer:
            self._output_projection = nn.Linear(in_features=num_heads * in_feats, out_features=out_feats)
        else:
            self._output_projection = nn.Linear(in_features=num_heads * in_feats, out_features=num_heads * out_feats)
        nn.init.xavier_normal_(self._output_projection.weight, gain=1.414)

    def forward(self, feat_dict, metapath_idx_dict):
        # Intra-metapath latent transformation
        feat_intra = {}
        for _metapath in self.metapath_list:
            feat_intra[_metapath] = \
                self.intra_metapath_trans(feat_dict, metapath=_metapath, metapath_idx_dict=metapath_idx_dict)

        # Inter-metapath latent transformation
        feat_inter = \
            self.inter_metapath_trans(feat_dict=feat_dict, feat_intra=feat_intra, metapath_list=self.metapath_list)

        # output projection
        feat_final = self.output_projection(feat_inter=feat_inter)

        # return final features after output projection (without nonlinear activation) and embedding
        # nonlinear activation will be added in MAGNN
        return feat_final, feat_inter

    def intra_metapath_trans(self, feat_dict, metapath, metapath_idx_dict):

        metapath_idx = metapath_idx_dict[metapath]

        # encoder metapath instances
        # intra_metapath_feat: feature matrix of every metapath instance of param metapath
        intra_metapath_feat = self.encoder(feat_dict, metapath, metapath_idx)

        # aggregate metapath instances into metapath using ATTENTION
        feat_intra = \
            self.intra_attn_layers[metapath]([intra_metapath_feat, feat_dict[metapath.split('-')[0]]],
                                             metapath, metapath_idx)
        return feat_intra

    def inter_metapath_trans(self, feat_dict, feat_intra, metapath_list):
        meta_s = {}
        feat_inter = {}
        # construct spi, where pi = ['M-A-M', 'M-D-M', ...]
        for metapath in metapath_list:
            _metapath = metapath.split('-')
            meta_feat = feat_intra[metapath]
            meta_feat = th.tanh(self.inter_linear[_metapath[0]](meta_feat)).mean(dim=0)  # s_pi
            meta_s[metapath] = self.inter_attn_vec[_metapath[0]](meta_feat)  # e_pi

        for ntype in self.ntypes:
            if ntype in self.dst_ntypes:
                # extract the metapath with the dst node type of ntype to construct a tensor
                # in order to compute softmax
                # metapaths: e.g if ntype is M, then ['M-A-M', 'M-D-M']
                metapaths = np.array(metapath_list)[[meta.split('-')[0] == ntype for meta in metapath_list]]
                # extract the e_pi of metapaths
                # e.g the e_pi of ['M-A-M', 'M-D-M'] if ntype is M
                meta_b = th.tensor(itemgetter(*metapaths)(meta_s))
                # compute softmax, obtain b_pi, which is attention score of metapaths
                # e.g the b_pi of ['M-A-M', 'M-D-M'] if ntype is M
                meta_b = F.softmax(meta_b, dim=0)
                # extract corresbonding features of metapath
                # e.g ['MDM_feat_attn', 'MAM_feat_attn'] if ntype is M
                meta_feat = itemgetter(*metapaths)(feat_intra)
                # compute the embedding feature of nodes
                feat_inter[ntype] = th.stack([meta_b[i] * meta_feat[i] for i in range(len(meta_b))], dim=0).sum(dim=0)
            else:
                feat_inter[ntype] = feat_dict[ntype]
        return feat_inter

    def encoder(self, feat_dict, metapath, metapath_idx):
        _metapath = metapath.split('-')
        device = feat_dict[_metapath[0]].device
        feat = th.zeros((len(_metapath), metapath_idx.shape[0], feat_dict[_metapath[0]].shape[1]), device=device)
        for i, ntype in zip(range(len(_metapath)), _metapath):
            feat[i] = feat_dict[ntype][metapath_idx[:, i]]
        feat = feat.reshape(feat.shape[0], feat.shape[1], feat.shape[2] // 2, 2)

        if self.encoder_type == 'RotateE':
            temp_r_vec = th.zeros((len(_metapath), feat.shape[-2], 2), device=device)
            temp_r_vec[0, :, 0] = 1

            for i in range(1, len(_metapath), 1):
                edge_type = '{}-{}'.format(_metapath[i - 1], _metapath[i])
                temp_r_vec[i] = self.complex_hada(temp_r_vec[i - 1], self.r_vec_dict[edge_type])
                feat[i] = self.complex_hada(feat[i], temp_r_vec[i], opt='feat')

            feat = feat.reshape(feat.shape[0], feat.shape[1], -1)
            return th.mean(feat, dim=0)

        elif self.encoder_type == 'Linear':
            feat = feat.reshape(feat.shape[0], feat.shape[1], -1)
            feat = self.encoder_linear(th.mean(feat, dim=0))
            return feat

        elif self.encoder_type == 'Average':
            feat = feat.reshape(feat.shape[0], feat.shape[1], -1)
            return th.mean(feat, dim=0)

        else:
            raise ValueError("The encoder type {} has not been implemented yet.".format(self.encoder_type))

    @staticmethod
    def complex_hada(h, v, opt='r_vec'):
        if opt == 'r_vec':
            h_h, l_h = h[:, 0].clone(), h[:, 1].clone()
        else:
            h_h, l_h = h[:, :, 0].clone(), h[:, :, 1].clone()
        h_v, l_v = v[:, 0].clone(), v[:, 1].clone()
        res = th.zeros_like(h)

        if opt == 'r_vec':
            res[:, 0] = h_h * h_v - l_h * l_v
            res[:, 1] = h_h * l_v + l_h * h_v
        else:
            res[:, :, 0] = h_h * h_v - l_h * l_v
            res[:, :, 1] = h_h * l_v + l_h * h_v
        return res

    # def output_projection(self, g):
    def output_projection(self, feat_inter):
        feat_final = {}
        for ntype in self.ntypes:
            feat_final[ntype] = self._output_projection(feat_inter[ntype])
        return feat_final


class MAGNN_attn_intra(nn.Module):
    def __init__(self, in_feats, out_feats, num_heads, feat_drop=0.5, attn_drop=0.5, negative_slope=0.01,
                 activation=F.elu):
        super(MAGNN_attn_intra, self).__init__()
        self._num_heads = num_heads
        self._in_src_feats, self._in_dst_feats = expand_as_pair(in_feats)
        self._out_feats = out_feats

        self.attn_r = nn.Parameter(th.FloatTensor(size=(1, num_heads, out_feats)))
        self.feat_drop = nn.Dropout(feat_drop)
        self.attn_drop = nn.Dropout(attn_drop)
        self.leaky_relu = nn.LeakyReLU(negative_slope)

        self.reset_parameters()
        self.activation = activation

    def reset_parameters(self):
        nn.init.xavier_normal_(self.attn_r, gain=1.414)

    def forward(self, feat, metapath, metapath_idx):
        _metapath = metapath.split('-')
        device = feat[0].device
        h_meta = self.feat_drop(feat[0]).view(-1, self._num_heads,
                                              self._out_feats)  # feature matrix of metapath instances

        # metapath(right) part of attention
        er = (h_meta * self.attn_r).sum(dim=-1).unsqueeze(-1)

        graph_data = {
            ('meta_inst', 'meta2{}'.format(_metapath[0]), _metapath[0]): (th.arange(0, metapath_idx.shape[0]),
                                                                          th.tensor(metapath_idx[:, 0]),)
        }
        num_nodes_dict = {'meta_inst': metapath_idx.shape[0], _metapath[0]: feat[1].shape[0]}

        g_meta = dgl.heterograph(graph_data, num_nodes_dict=num_nodes_dict).to(device)

        # feature vector of metapath instances and nodes
        g_meta.nodes['meta_inst'].data.update({'feat_src': h_meta, 'er': er})
        # g_meta.nodes[metapath[0]].data.update({'feat':feat[1]})

        # compute attention without concat with hv
        g_meta.apply_edges(func=fn.copy_u('er', 'e'), etype='meta2{}'.format(_metapath[0]))

        e = self.leaky_relu(g_meta.edata.pop('e'))
        g_meta.edata['a'] = self.attn_drop(edge_softmax(g_meta, e))

        # message passing, there's only one edge type
        # by default DGL would fill nodes without in-degree with zero
        g_meta.update_all(message_func=fn.u_mul_e('feat_src', 'a', 'm'), reduce_func=fn.sum('m', 'feat'))

        feat = self.activation(g_meta.dstdata['feat'])

        # return dst nodes' features after attention
        return feat.flatten(1)


'''
methods
'''


def mp_instance_sampler(g, metapath_list, dataset):
    """
    Sampling the indices of all metapath instances in g according to the metapath list

    Parameters
    ----------
    g : object
        the dgl heterogeneous graph
    metapath_list : list
        the list of metapaths in g, e.g. ['M-A-M', M-D-M', ...]
    dataset : str
        the name of dataset, e.g. 'imdb4MAGNN'

    Returns
    -------
    dict
        the indices of all metapath instances. e.g dict['MAM'] contains the indices of all MAM instances

    Notes
    -----
    Please make sure that the metapath in metapath_list are all symmetric

    We'd store the metapath instances in the disk after one metapath instances sampling and next time the
    metapath instances will be extracted directly from the disk if they exists.

    """

    file_dir = 'openhgnn/output/MAGNN/'
    file_addr = file_dir + '{}'.format(dataset) + '_mp_inst.pkl'
    test = True  # TODO

    if os.path.exists(file_addr) and test is False:  # TODO
        with open(file_addr, 'rb') as file:
            res = pickle.load(file)
    else:
        etype_idx_dict = {}
        for etype in g.etypes:
            edges_idx_i = g.edges(etype=etype)[0].cpu().numpy()
            edges_idx_j = g.edges(etype=etype)[1].cpu().numpy()
            etype_idx_dict[etype] = pd.DataFrame([edges_idx_i, edges_idx_j]).T
            _etype = etype.split('-')
            etype_idx_dict[etype].columns = [_etype[0], _etype[1]]

        res = {}
        for metapath in metapath_list:
            res[metapath] = None
            _metapath = metapath.split('-')
            for i in range(1, len(_metapath) - 1):
                if i == 1:
                    res[metapath] = etype_idx_dict['-'.join(_metapath[:i + 1])]
                feat_j = etype_idx_dict['-'.join(_metapath[i:i + 2])]
                col_i = res[metapath].columns[-1]
                col_j = feat_j.columns[0]
                res[metapath] = pd.merge(res[metapath], feat_j,
                                         left_on=col_i,
                                         right_on=col_j,
                                         how='inner')
                if col_i != col_j:
                    res[metapath].drop(columns=col_j, inplace=True)
            res[metapath] = res[metapath].values

        with open(file_addr, 'wb') as file:
            pickle.dump(res, file)

    return res


def mini_mp_instance_sampler(seed_nodes, mp_instances, num_samples):
    '''
    Sampling metapath instances with seed_nodes as dst nodes. This method is exclusive to mini batch train/validate/test
    which need to sample subsets of metapath instances of the whole graph.

    Parameters
    ----------
    seed_nodes : dict
        sampling metapath instances based on seed_nodes. e.g. {'A':[0, 1, 2], 'M':[0, 1, 2], ...}, then we'll sample
        metapath instances with 0 or 1 or 2 as dst_nodes of type 'A' and type 'B'.
    mp_instances : list
        the sampled metapath instances of the whole graph. It should be the return value of method
        ``mp_instance_sampler(g, metapath_list, dataset)``
    num_samples : int
        the maximal number of sampled metapath instances of each metapath type.

    Returns
    -------
    dict
        sampled metapath instances

    '''
    mini_mp_inst = {}
    metapath_list = list(mp_instances.keys())

    for ntype in seed_nodes.keys():
        target_mp_types = np.array(metapath_list)[[meta.split('-')[0] == ntype for meta in metapath_list]]
        for metapath in target_mp_types:  # the metapath instances of the certain metapath
            for node in seed_nodes[ntype]:
                _mp_inst = mp_instances[metapath][mp_instances[metapath][:, 0] == node]
                dst_nodes, dst_counts = np.unique(_mp_inst[:, -1], return_counts=True)

                # the method of computing sampling probabilities originates from author's codes
                p = np.repeat((dst_counts ** (3 / 4)) / dst_counts, dst_counts)
                p = p / p.sum()

                _num_samples = min(num_samples, len(p))
                mp_choice = np.random.choice(len(p), _num_samples, replace=False, p=p)
                if metapath not in mini_mp_inst.keys():
                    mini_mp_inst[metapath] = _mp_inst[mp_choice]
                else:
                    mini_mp_inst[metapath] = np.concatenate((mini_mp_inst[metapath], _mp_inst[mp_choice]),
                                                            axis=0)

    return mini_mp_inst


def svm_test(X, y, test_sizes=(0.2, 0.4, 0.6, 0.8), repeat=10):
    # This method is implemented by author
    random_states = [182318 + i for i in range(repeat)]
    result_macro_f1_list = []
    result_micro_f1_list = []
    for test_size in test_sizes:
        macro_f1_list = []
        micro_f1_list = []
        for i in range(repeat):
            X_train, X_test, y_train, y_test = train_test_split(
                X, y, test_size=test_size, shuffle=True, random_state=random_states[i])
            svm = LinearSVC(dual=False)
            svm.fit(X_train, y_train)
            y_pred = svm.predict(X_test)
            macro_f1 = f1_score(y_test, y_pred, average='macro')
            micro_f1 = f1_score(y_test, y_pred, average='micro')
            macro_f1_list.append(macro_f1)
            micro_f1_list.append(micro_f1)
        result_macro_f1_list.append((np.mean(macro_f1_list), np.std(macro_f1_list)))
        result_micro_f1_list.append((np.mean(micro_f1_list), np.std(micro_f1_list)))
    return result_macro_f1_list, result_micro_f1_list



import pickle
import torch as th
import numpy as np
import dgl
import os
from dgl.data import DGLBuiltinDataset
from dgl.data.utils import idx2mask, load_graphs, save_graphs

__all__ = ['GTNDataset', 'IMDB4GTNDataset', 'ACM4GTNDataset', 'DBLP4GTNDataset']


class GTNDataset(DGLBuiltinDataset):
    r"""GTN Dataset.

    It contains three datasets used in a NeurIPS'19 paper Graph Transformer Networks <https://arxiv.org/abs/1911.06455>,
    which includes two citation network datasets DBLP and ACM, and a movie dataset
    IMDB. DBLP contains three types of nodes (papers (P), authors (A), conferences (C)), four types of edges
    (PA, AP, PC, CP), and research areas of authors as labels. ACM contains three types of nodes (papers
    (P), authors (A), subject (S)), four types of edges (PA, AP, PS, SP), and categories of papers as labels.
    Each node in the two datasets is represented as bag-of-words of keywords. On the other hand, IMDB
    contains three types of nodes (movies (M), actors (A), and directors (D)) and labels are genres of
    movies. Node features are given as bag-of-words representations of plots.

    Dataset statistics:

    Dataset Nodes Edges Edge type Features Training Validation Test

    DBLP 18405 67946 4 334 800 400 2857

    ACM 8994 25922 4 1902 600 300 2125

    IMDB 12772 37288 4 1256 300 300 2339

    Data source link: <https://drive.google.com/file/d/1qOZ3QjqWMIIvWjzrIdRe3EA4iKzPi6S5/view?usp=sharing>

    Parameters
    ----------
    name : str
        Name of the dataset. Supported dataset names are 'dblp4GTN', 'acm4GTN' and 'imdb4GTN'.
    raw_dir : str
        Specifying the directory that will store the
        downloaded data or the directory that
        already stores the input data.
        Default: ~/.dgl/
    force_reload : bool
        Whether to reload the dataset. Default: False
    verbose : bool
        Whether to print out progress information. Default: False
    transform : callable, optional
        A transform that takes in a :class:`~dgl.DGLGraph` object and returns
        a transformed version. The :class:`~dgl.DGLGraph` object will be
        transformed before every access.

    Examples
    --------
    >>> dataset = GTNDataset(name='imdb4GTN')
    >>> graph = dataset[0]
    """

    def __init__(self, name, raw_dir=None, force_reload=False, verbose=False,
                 transform=None):
        assert name in ['dblp4GTN', 'acm4GTN', 'imdb4GTN']
        if name == 'dblp4GTN':
            canonical_etypes = [('paper', 'paper-author', 'author'), ('author', 'author-paper', 'paper'),
                                ('paper', 'paper-conference', 'conference'),
                                ('conference', 'conference-paper', 'paper')]
            target_ntype = 'author'
            meta_paths_dict = {'APCPA': [('author', 'author-paper', 'paper'),
                                         ('paper', 'paper-conference', 'conference'),
                                         ('conference', 'conference-paper', 'paper'),
                                         ('paper', 'paper-author', 'author')],
                               'APA': [('author', 'author-paper', 'paper'),
                                       ('paper', 'paper-author', 'author')],
                               }

        elif name == 'acm4GTN':
            canonical_etypes = [('paper', 'paper-author', 'author'), ('author', 'author-paper', 'paper'),
                                ('paper', 'paper-subject', 'subject'), ('subject', 'subject-paper', 'paper')]
            target_ntype = 'paper'
            meta_paths_dict = {'PAPSP': [('paper', 'paper-author', 'author'),
                                         ('author', 'author-paper', 'paper'),
                                         ('paper', 'paper-subject', 'subject'),
                                         ('subject', 'subject-paper', 'paper')],
                               'PAP': [('paper', 'paper-author', 'author'),
                                       ('author', 'author-paper', 'paper')],
                               'PSP': [('paper', 'paper-subject', 'subject'),
                                       ('subject', 'subject-paper', 'paper')]
                               }
        elif name == 'imdb4GTN':
            canonical_etypes = [('movie', 'movie-director', 'director'), ('director', 'director-movie', 'movie'),
                                ('movie', 'movie-actor', 'actor'), ('actor', 'actor-movie', 'movie')]
            target_ntype = 'movie'
            meta_paths_dict = {'MAM': [('movie', 'movie-actor', 'actor'),
                                       ('actor', 'actor-movie', 'movie')],
                               'MDM': [('movie', 'movie-director', 'director'),
                                       ('director', 'director-movie', 'movie')]
                               }
        else:
            raise ValueError('Unsupported dataset name {}'.format(name))
        self._canonical_etypes = canonical_etypes
        self._target_ntype = target_ntype
        self._meta_paths_dict = meta_paths_dict
        super(GTNDataset, self).__init__(
            name,
            url='https://s3.cn-north-1.amazonaws.com.cn/dgl-data/dataset/openhgnn/{}.zip'.format(name),
            raw_dir=raw_dir,
            force_reload=force_reload, verbose=verbose, transform=transform)

    def process(self):
        target_ntype = self.target_ntype
        canonical_etypes = self._canonical_etypes

        if os.path.isfile(os.path.join(self.save_path, 'graph.bin')):# Has cache
            graph_path = os.path.join(self.save_path, 'graph.bin')
            gs, _ = load_graphs(graph_path)
            g = gs[0]
        else:
            with open(self.raw_path + '/node_features.pkl', 'rb') as f:
                node_features = pickle.load(f)
            with open(self.raw_path + '/edges.pkl', 'rb') as f:
                edges = pickle.load(f)
            with open(self.raw_path + '/labels.pkl', 'rb') as f:
                labels = pickle.load(f)

            num_nodes = edges[0].shape[0]
            assert len(canonical_etypes) == len(edges)

            ntype_mask = dict()
            ntype_idmap = dict()
            ntypes = set()
            data_dict = {}

            # create dgl graph
            for etype in canonical_etypes:
                ntypes.add(etype[0])
                ntypes.add(etype[2])
            for ntype in ntypes:
                ntype_mask[ntype] = np.zeros(num_nodes, dtype=bool)
                ntype_idmap[ntype] = np.full(num_nodes, -1, dtype=int)
            for i, etype in enumerate(canonical_etypes):
                src_nodes = edges[i].nonzero()[0]
                dst_nodes = edges[i].nonzero()[1]
                src_ntype = etype[0]
                dst_ntype = etype[2]
                ntype_mask[src_ntype][src_nodes] = True
                ntype_mask[dst_ntype][dst_nodes] = True
            for ntype in ntypes:
                ntype_idx = ntype_mask[ntype].nonzero()[0]
                ntype_idmap[ntype][ntype_idx] = np.arange(ntype_idx.size)
            for i, etype in enumerate(canonical_etypes):
                src_nodes = edges[i].nonzero()[0]
                dst_nodes = edges[i].nonzero()[1]
                src_ntype = etype[0]
                dst_ntype = etype[2]
                data_dict[etype] = \
                    (th.from_numpy(ntype_idmap[src_ntype][src_nodes]).type(th.int64),
                    th.from_numpy(ntype_idmap[dst_ntype][dst_nodes]).type(th.int64))
            g = dgl.heterograph(data_dict)

            # split and label
            all_label = np.full(g.num_nodes(target_ntype), -1, dtype=int)
            for i, split in enumerate(['train', 'val', 'test']):
                node = np.array(labels[i])[:, 0]
                label = np.array(labels[i])[:, 1]
                all_label[node] = label
                g.nodes[target_ntype].data['{}_mask'.format(split)] = \
                    th.from_numpy(idx2mask(node, g.num_nodes(target_ntype))).type(th.bool)
            g.nodes[target_ntype].data['label'] = th.from_numpy(all_label).type(th.long)

            # node feature
            node_features = th.from_numpy(node_features).type(th.FloatTensor)
            for ntype in ntypes:
                idx = ntype_mask[ntype].nonzero()[0]
                g.nodes[ntype].data['h'] = node_features[idx]

        self._g = g
        self._num_classes = len(th.unique(self._g.nodes[self.target_ntype].data['label']))
        self._in_dim = self._g.ndata['h'][self.target_ntype].shape[1]

    def save(self):
        graph_path = os.path.join(self.save_path, 'graph.bin')
        save_graphs(graph_path, self._g)

    def load(self):
        graph_path = os.path.join(self.save_path, 'graph.bin')
        gs, _ = load_graphs(graph_path)
        self._g = gs[0]
        self._num_classes = len(th.unique(self._g.nodes[self.target_ntype].data['label']))
        self._in_dim = self._g.ndata['h'][self.target_ntype].shape[1]
    
    def has_cache(self):
        return os.path.isfile(os.path.join(self.save_path, 'graph.bin'))

    @property
    def target_ntype(self):
        return self._target_ntype

    @property
    def category(self):
        return self._target_ntype

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def meta_paths_dict(self):
        return self._meta_paths_dict

    @property
    def in_dim(self):
        return self._in_dim

    def __getitem__(self, idx):
        assert idx == 0
        return self._g

    def __len__(self):
        return 1


class DBLP4GTNDataset(GTNDataset):
    def __init__(self, raw_dir=None, force_reload=False, verbose=False, transform=None):
        name = 'dblp4GTN'
        super(DBLP4GTNDataset, self).__init__(name, raw_dir=raw_dir, force_reload=force_reload, verbose=verbose,
                                              transform=transform)


class ACM4GTNDataset(GTNDataset):
    def __init__(self, raw_dir=None, force_reload=False, verbose=False, transform=None):
        name = 'acm4GTN'
        super(ACM4GTNDataset, self).__init__(name, raw_dir=raw_dir, force_reload=force_reload, verbose=verbose,
                                             transform=transform)


class IMDB4GTNDataset(GTNDataset):
    def __init__(self, raw_dir=None, force_reload=False, verbose=False, transform=None):
        name = 'imdb4GTN'
        super(IMDB4GTNDataset, self).__init__(name, raw_dir=raw_dir, force_reload=force_reload, verbose=verbose,
                                              transform=transform)



from .EmbedLayer import HeteroEmbedLayer
from .HeteroLinear import GeneralLinear, HeteroLinearLayer
from .HeteroLinear import HeteroMLPLayer, HeteroFeature
from .MetapathConv import MetapathConv
from .HeteroGraphConv import HeteroGraphConv
from .macro_layer import *
from .micro_layer import *

__all__ = [
    'HeteroEmbedLayer',
    'GeneralLinear',
    'HeteroLinearLayer',
    'HeteroMLPLayer',
    'HeteroFeature',
    'MetapathConv',
    'HeteroGraphConv',
    'ATTConv',
    'MacroConv',
    'SemanticAttention',
    'CompConv',
    'AttConv',
    'LSTMConv'
]

classes = __all__


import math
import dgl
import torch
import torch.nn as nn

from dgl import function as fn
from dgl.nn.pytorch.linear import TypedLinear
from dgl.nn.pytorch.softmax import edge_softmax
from . import BaseModel, register_model
from ..utils import to_hetero_feat


@register_model('HGT')
class HGT(BaseModel):
    r"""Heterogeneous graph transformer convolution from `Heterogeneous Graph Transformer
    <https://arxiv.org/abs/2003.01332>`__

    For more details, you may refer to `HGT<https://docs.dgl.ai/en/0.8.x/generated/dgl.nn.pytorch.conv.HGTConv.html>`__
    
    Parameters
    ----------
    in_dim: int
        the input dimension
    out_dim: int
        the output dimension
    num_heads: list
        the list of the number of heads in each layer
    num_etypes: int
        the number of the edge type
    num_ntypes: int
        the number of the node type
    num_layers: int
        the number of layers we used in the computing
    dropout: float
        the feature drop rate
    norm: boolean
        if we need the norm operation
    ntypes: list
        the list of node type
    """

    @classmethod
    def build_model_from_args(cls, args, hg):

        return cls(args.hidden_dim,
                   args.out_dim,
                   args.num_heads,
                   len(hg.etypes),
                   hg.ntypes,
                   args.num_layers,
                   args.dropout,
                   args.norm
                   )

    def __init__(self, in_dim, out_dim, num_heads, num_etypes, ntypes,
                 num_layers, dropout=0.2, norm=False):
        super(HGT, self).__init__()
        self.num_layers = num_layers
        self.ntypes = ntypes
        self.hgt_layers = nn.ModuleList()
        self.hgt_layers.append(
            HGTConv(in_dim,
                    in_dim // num_heads,
                    num_heads,
                    len(ntypes),
                    num_etypes,
                    dropout,
                    norm)
        )

        for _ in range(1, num_layers - 1):
            self.hgt_layers.append(
                HGTConv(in_dim,
                        in_dim // num_heads,
                        num_heads,
                        len(ntypes),
                        num_etypes,
                        dropout,
                        norm)
            )

        self.hgt_layers.append(
            HGTConv(in_dim,
                    out_dim,
                    1,
                    len(ntypes),
                    num_etypes,
                    dropout,
                    norm)
        )

    def forward(self, hg, h_dict):
        """
        The forward part of the HGT.
        
        Parameters
        ----------
        hg : object
            the dgl heterogeneous graph
        h_dict: dict
            the feature dict of different node types
            
        Returns
        -------
        dict
            The embeddings after the output projection.
        """

        if hasattr(hg, 'ntypes'):
            # full graph training,
            with hg.local_scope():
                hg.ndata['h'] = h_dict
                g = dgl.to_homogeneous(hg, ndata='h')
                h = g.ndata['h']
                for l in range(self.num_layers):
                    h = self.hgt_layers[l](g, h, g.ndata['_TYPE'], g.edata['_TYPE'], presorted=True)
            h_dict = to_hetero_feat(h, g.ndata['_TYPE'], self.ntypes)

        else:
            # for minibatch training, input h_dict is a tensor
            h = h_dict
            for layer, block in zip(self.hgt_layers, hg):
                h = layer(block, h, block.ndata['_TYPE']['_N'], block.edata['_TYPE'], presorted=False)
            h_dict = to_hetero_feat(h, block.ndata['_TYPE']['_N'][:block.num_dst_nodes()], self.ntypes)

        return h_dict

    @property
    def to_homo_flag(self):
        return True

class HGTConv(nn.Module):
    r"""Heterogeneous graph transformer convolution from `Heterogeneous Graph Transformer
    <https://arxiv.org/abs/2003.01332>`__

    Given a graph :math:`G(V, E)` and input node features :math:`H^{(l-1)}`,
    it computes the new node features as follows:

    Compute a multi-head attention score for each edge :math:`(s, e, t)` in the graph:

    .. math::

      Attention(s, e, t) = \text{Softmax}\left(||_{i\in[1,h]}ATT-head^i(s, e, t)\right) \\
      ATT-head^i(s, e, t) = \left(K^i(s)W^{ATT}_{\phi(e)}Q^i(t)^{\top}\right)\cdot
        \frac{\mu_{(\tau(s),\phi(e),\tau(t)}}{\sqrt{d}} \\
      K^i(s) = \text{K-Linear}^i_{\tau(s)}(H^{(l-1)}[s]) \\
      Q^i(t) = \text{Q-Linear}^i_{\tau(t)}(H^{(l-1)}[t]) \\

    Compute the message to send on each edge :math:`(s, e, t)`:

    .. math::

      Message(s, e, t) = ||_{i\in[1, h]} MSG-head^i(s, e, t) \\
      MSG-head^i(s, e, t) = \text{M-Linear}^i_{\tau(s)}(H^{(l-1)}[s])W^{MSG}_{\phi(e)} \\

    Send messages to target nodes :math:`t` and aggregate:

    .. math::

      \tilde{H}^{(l)}[t] = \sum_{\forall s\in \mathcal{N}(t)}\left( Attention(s,e,t)
      \cdot Message(s,e,t)\right)

    Compute new node features:

    .. math::

      H^{(l)}[t]=\text{A-Linear}_{\tau(t)}(\sigma(\tilde(H)^{(l)}[t])) + H^{(l-1)}[t]

    Parameters
    ----------
    in_size : int
        Input node feature size.
    head_size : int
        Output head size. The output node feature size is ``head_size * num_heads``.
    num_heads : int
        Number of heads. The output node feature size is ``head_size * num_heads``.
    num_ntypes : int
        Number of node types.
    num_etypes : int
        Number of edge types.
    dropout : optional, float
        Dropout rate.
    use_norm : optiona, bool
        If true, apply a layer norm on the output node feature.

    Examples
    --------
    """
    def __init__(self,
                 in_size,
                 head_size,
                 num_heads,
                 num_ntypes,
                 num_etypes,
                 dropout=0.2,
                 use_norm=False):
        super().__init__()
        self.in_size = in_size
        self.head_size = head_size
        self.num_heads = num_heads
        self.sqrt_d = math.sqrt(head_size)
        self.use_norm = use_norm

        self.linear_k = TypedLinear(in_size, head_size * num_heads, num_ntypes)
        self.linear_q = TypedLinear(in_size, head_size * num_heads, num_ntypes)
        self.linear_v = TypedLinear(in_size, head_size * num_heads, num_ntypes)
        self.linear_a = TypedLinear(head_size * num_heads, head_size * num_heads, num_ntypes)

        self.relation_pri = nn.ParameterList([nn.Parameter(torch.ones(num_etypes))
                                              for i in range(num_heads)])
        self.relation_att = nn.ModuleList([TypedLinear(head_size, head_size, num_etypes)
                                           for i in range(num_heads)])
        self.relation_msg = nn.ModuleList([TypedLinear(head_size, head_size, num_etypes)
                                           for i in range(num_heads)])
        self.skip = nn.Parameter(torch.ones(num_ntypes))
        self.drop = nn.Dropout(dropout)
        if use_norm:
            self.norm = nn.LayerNorm(head_size * num_heads)
        if in_size != head_size * num_heads:
            self.residual_w = nn.Parameter(torch.Tensor(in_size, head_size * num_heads))
            nn.init.xavier_uniform_(self.residual_w)

    def forward(self, g, x, ntype, etype, *, presorted=False):
        """Forward computation.

        Parameters
        ----------
        g : DGLGraph
            The input graph.
        x : torch.Tensor
            A 2D tensor of node features. Shape: :math:`(|V|, D_{in})`.
        ntype : torch.Tensor
            An 1D integer tensor of node types. Shape: :math:`(|V|,)`.
        etype : torch.Tensor
            An 1D integer tensor of edge types. Shape: :math:`(|E|,)`.
        presorted : bool, optional
            Whether *both* the nodes and the edges of the input graph have been sorted by
            their types. Forward on pre-sorted graph may be faster. Graphs created by
            :func:`~dgl.to_homogeneous` automatically satisfy the condition.
            Also see :func:`~dgl.reorder_graph` for manually reordering the nodes and edges.

        Returns
        -------
        torch.Tensor
            New node features. Shape: :math:`(|V|, D_{head} * N_{head})`.
        """
        self.presorted = presorted
        if g.is_block:
            x_src = x
            x_dst = x[:g.num_dst_nodes()]
            srcntype = ntype
            dstntype = ntype[:g.num_dst_nodes()]
        else:
            x_src = x
            x_dst = x
            srcntype = ntype
            dstntype = ntype
        with g.local_scope():
            k = self.linear_k(x_src, srcntype, presorted).view(-1, self.num_heads, self.head_size)
            q = self.linear_q(x_dst, dstntype, presorted).view(-1, self.num_heads, self.head_size)
            v = self.linear_v(x_src, srcntype, presorted).view(-1, self.num_heads, self.head_size)
            g.srcdata['k'] = k
            g.dstdata['q'] = q
            g.srcdata['v'] = v
            g.edata['etype'] = etype
            g.apply_edges(self.message)
            g.edata['m'] = g.edata['m'] * edge_softmax(g, g.edata['a']).unsqueeze(-1)
            g.update_all(fn.copy_e('m', 'm'), fn.sum('m', 'h'))
            h = g.dstdata['h'].view(-1, self.num_heads * self.head_size)
            # target-specific aggregation
            h = self.drop(self.linear_a(h, dstntype, presorted))
            alpha = torch.sigmoid(self.skip[dstntype]).unsqueeze(-1)
            if x_dst.shape != h.shape:
                h = h * alpha + (x_dst @ self.residual_w) * (1 - alpha)
            else:
                h = h * alpha + x_dst * (1 - alpha)
            if self.use_norm:
                h = self.norm(h)
            return h

    def message(self, edges):
        """Message function."""
        a, m = [], []
        etype = edges.data['etype']
        k = torch.unbind(edges.src['k'], dim=1)
        q = torch.unbind(edges.dst['q'], dim=1)
        v = torch.unbind(edges.src['v'], dim=1)
        for i in range(self.num_heads):
            kw = self.relation_att[i](k[i], etype, self.presorted)  # (E, O)
            a.append((kw * q[i]).sum(-1) * self.relation_pri[i][etype] / self.sqrt_d)   # (E,)
            m.append(self.relation_msg[i](v[i], etype, self.presorted))  # (E, O)
        return {'a' : torch.stack(a, dim=1), 'm' : torch.stack(m, dim=1)}



import numpy
import os
import json

import numpy as np
import torch as th
import torch.sparse as sparse
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
from . import BaseModel, register_model
import dgl

@register_model('GIE')
class GIE(BaseModel):
    @classmethod
    def build_model_from_args(cls, args, hg):
        return cls(args, hg)
    def __init__(self, args, hg):
        super(GIE, self).__init__()
        self.device = args.device
        self.ent_num = hg.num_nodes()
        self.rel_num = len(hg.canonical_etypes)
        self.dim = args.hidden_dim
        self.margin = args.margin
        self.dis_norm = args.dis_norm
        self.scale = th.Tensor([1./np.sqrt(self.dim)]).double().to(self.device)
        self.n_emb = nn.Embedding(self.ent_num, self.dim)
        self.r_emb = nn.Embedding(self.rel_num, self.dim)
        self.r_diagE = nn.Embedding(self.rel_num, self.dim)
        self.r_diagH = nn.Embedding(self.rel_num, self.dim)
        self.r_diagS = nn.Embedding(self.rel_num, self.dim)
        self.context_vec = nn.Embedding(self.rel_num, self.dim)
        self.act = nn.Softmax(dim=1)
        self.c = nn.Parameter(th.ones((self.rel_num, 1), dtype=th.double), requires_grad=True)
        self.v = nn.Parameter(th.ones((self.rel_num, 1), dtype=th.double), requires_grad=True)
        self.u = nn.Parameter(th.ones((self.rel_num, 1), dtype=th.double), requires_grad=True)
        self.MIN_NORM = 1e-15
        nn.init.xavier_uniform_(self.n_emb.weight.data)
        nn.init.xavier_uniform_(self.r_emb.weight.data)
        nn.init.xavier_uniform_(self.r_diagE.weight.data)
        nn.init.xavier_uniform_(self.r_diagH.weight.data)
        nn.init.xavier_uniform_(self.r_diagS.weight.data)
        nn.init.xavier_uniform_(self.context_vec.weight.data)
    def forward(self, h, r, t):
        h = h.to(self.device)
        r = r.to(self.device)
        t = t.to(self.device)
        if h.shape == th.Size([]):
            h = h.view(1).repeat(t.shape[0])
        if r.shape == th.Size([]):
            r = r.view(1).repeat(h.shape[0])
        if t.shape == th.Size([]):
            t = t.view(1).repeat(h.shape[0])
        h_emb = self.n_emb(h)
        r_emb = self.r_emb(r)
        t_emb = self.n_emb(t)
        ## Euclidean space projection
        c = F.softplus(self.c[r])
        r_exp0c = self.expmap0(r_emb, c)
        res_E = self.givens_rotations(self.r_diagE(r), h_emb).view((-1, 1, self.dim))

        ## hyperbolic space projection
        v = F.softplus(self.v[r])
        h_exp0v = self.expmap0(h_emb, v)
        r_exp0v = self.expmap0(r_emb, v)
        lh_H = self.project(self.mobius_add(h_exp0v, r_exp0v, v), v)
        res_H = self.logmap0(self.givens_rotations(self.r_diagH(r), lh_H), v).view((-1, 1, self.dim))

        ## Hyperspherical Space Projection
        u = F.softplus(self.u[r])
        h_exp0u = self.expmap0(h_emb, u)
        r_exp0u = self.expmap0(r_emb, u)
        lh_S = self.project(self.mobius_add(h_exp0u, r_exp0u, u), u)
        res_S = self.logmap0(self.givens_rotations(self.r_diagS(r), lh_S), u).view((-1, 1, self.dim))

        ## Aggregation with weights
        cands = th.cat([res_E, res_H, res_S], dim=1)
        context_vec = self.context_vec(r).view((-1, 1, self.dim))
        att_W = self.act(th.sum(context_vec * cands * self.scale, dim=-1, keepdim=True))
        ## Inter(E_h, H_h, S_h) = exp_0^c(\lambda_E E_h + \lambda_H log_0^v(H_h) + \lambda_S log_0^u(S_h))
        hr_emb = self.project(self.mobius_add(self.expmap0(th.sum(att_W * cands, dim=1), c), r_exp0c, c),c)


        return (self.similarity_score(hr_emb, t_emb, c) + self.margin).view(-1)

    def similarity_score(self, x, v, c):
        sqrt_c = c ** 0.5
        v_norm = th.norm(v, p=2, dim=-1, keepdim=True)
        xv = th.sum(x * v / v_norm, dim=-1, keepdim=True)
        gamma = self.tanh(sqrt_c * v_norm) / sqrt_c
        x2 = th.sum(x * x, dim=-1, keepdim=True)
        c1 = 1 - 2 * c * gamma * xv + c * gamma ** 2
        c2 = 1 - c * x2
        num = th.sqrt((c1 ** 2) * x2 + (c2 ** 2) * (gamma ** 2) - (2 * c1 * c2) * gamma * xv)
        denom = 1 - 2 * c * gamma * xv + (c ** 2) * (gamma ** 2) * x2
        pairwise_norm = num / denom.clamp_min(self.MIN_NORM)
        dist = self.artanh(sqrt_c * pairwise_norm)
        return - (2 * dist / sqrt_c) ** 2

    def tanh(self, x):
        return x.clamp(-15, 15).tanh()

    def artanh(self, x):
        return Artanh.apply(x)

    def givens_rotations(self, r, x):
        ## vector rotation
        givens = r.view((r.shape[0], -1, 2))
        givens = givens / th.norm(givens, p=2, dim=-1, keepdim=True).clamp_min(1e-15)
        x = x.view((r.shape[0], -1, 2))
        x_rot = givens[:, :, 0:1] * x + givens[:, :, 1:] * th.cat((-x[:, :, 1:], x[:, :, 0:1]), dim=-1)
        return x_rot.view((r.shape[0], -1))

    def mobius_add(self, x, y, c):
        ## Möbius addition
        x2 = th.sum(x * x, dim=-1, keepdim=True)
        y2 = th.sum(y * y, dim=-1, keepdim=True)
        xy = th.sum(x * y, dim=-1, keepdim=True)
        num = (1 + 2 * c * xy + c * y2) * x + (1 - c * x2) * y
        denom = 1 + 2 * c * xy + c ** 2 * x2 * y2
        return num / denom.clamp_min(self.MIN_NORM)

    def expmap0(self, u, c):
        sqrt_c = c ** 0.5
        u_norm = u.norm(dim=-1, p=2, keepdim=True).clamp_min(self.MIN_NORM)
        ## tanh(\sqrt{c} {\lambda_x ||u||}/2)*v/\sqrt{c} {\lambda_x ||u||}/2
        temp = self.tanh(sqrt_c * u_norm) * u
        gamma_1 = temp / (sqrt_c * u_norm)
        return self.project(gamma_1, c)

    def logmap0(self, y, c):
        sqrt_c = c ** 0.5
        y_norm = y.norm(dim=-1, p=2, keepdim=True).clamp_min(self.MIN_NORM)
        return y / y_norm / sqrt_c * self.artanh(sqrt_c * y_norm)

    def project(self, x, c):
        norm = x.norm(dim=-1, p=2, keepdim=True).clamp_min(self.MIN_NORM)
        eps = 1e-5
        ## \frac{1}{\sqrt{c}}
        maxnorm = (1 - eps) / (c ** 0.5)
        cond = norm > maxnorm
        projected = x / norm * maxnorm
        return th.where(cond, projected, x)

class Artanh(th.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        x = x.clamp(-1 + 1e-5, 1 - 1e-5)
        ctx.save_for_backward(x)
        dtype = x.dtype
        x = x.double()
        return (th.log_(1 + x).sub_(th.log_(1 - x))).mul_(0.5).to(dtype)

    @staticmethod
    def backward(ctx, grad_output):
        input, = ctx.saved_tensors
        return grad_output / (1 - input ** 2)




import os
from dgl.data.utils import download, extract_archive
from dgl.data import DGLDataset
from dgl.data.utils import load_graphs, idx2mask
import pickle
import scipy
import numpy as np
import dgl
import torch as th


class AcademicDataset(DGLDataset):

    _prefix = 'https://s3.cn-north-1.amazonaws.com.cn/dgl-data/'
    _urls = {
        'academic4HetGNN': 'dataset/academic4HetGNN.zip',
        'acm4GTN': 'dataset/acm4GTN.zip',
        'acm4NSHE': 'dataset/acm4NSHE.zip',
        'acm4NARS': 'dataset/acm4NARS.zip',
        'acm4HeCo': 'dataset/acm4HeCo.zip',
        'imdb4MAGNN': 'dataset/imdb4MAGNN.zip',
        'imdb4GTN': 'dataset/imdb4GTN.zip',
        'DoubanMovie': 'dataset/DoubanMovie.zip',
        'dblp4MAGNN': 'dataset/dblp4MAGNN.zip',
        'yelp4HeGAN': 'dataset/yelp4HeGAN.zip',
        'yelp4rec': 'dataset/yelp4rec.zip',
        'HNE-PubMed': 'dataset/HNE-PubMed.zip',
        'MTWM': 'dataset/MTWM3.zip',
        'amazon4SLICE': 'dataset/amazon4SLICE.zip',
        'amazon': 'https://zhiguli.oss-cn-hangzhou.aliyuncs.com/amazon.zip',
        'yelp4HGSL': 'dataset/yelp4HGSL.zip'
    }

    def __init__(self, name, raw_dir=None, force_reload=False, verbose=True):
        assert name in ['acm4GTN', 'acm4NSHE', 'academic4HetGNN', 'imdb4MAGNN', 'imdb4GTN', 'HNE-PubMed', 'MTWM',
                        'DoubanMovie', 'dblp4MAGNN', 'acm4NARS', 'acm4HeCo', 'yelp4rec', 'yelp4HeGAN', 'amazon4SLICE','amazon', 'yelp4HGSL']
        if name == 'yelp4HGSL':
            canonical_etypes = [('b', 'b-s', 's'), ('s', 's-b', 'b'), ('b', 'b-l', 'l'), ('l', 'l-b', 'b'), ('b', 'b-u', 'u'),
                                ('u', 'u-b', 'b')]
            target_ntype = 'b'
            meta_paths_dict = {'bub': [('b', 'b-u', 'u'), ('u', 'u-b', 'b')],
                            'bsb': [('b', 'b-s', 's'), ('s', 's-b', 'b')],
                            'bublb': [('b', 'b-u', 'u'), ('u', 'u-b', 'b'),
                                        ('b', 'b-l', 'l'), ('l', 'l-b', 'b')],
                            'bubsb': [('b', 'b-u', 'u'), ('u', 'u-b', 'b'),
                                        ('b', 'b-s', 's'), ('s', 's-b', 'b')]
                   }
            self._canonical_etypes = canonical_etypes
            self._target_ntype = target_ntype
            self._meta_paths_dict = meta_paths_dict
        self.data_path = './openhgnn/' + self._urls[name]
        self.g_path = './openhgnn/dataset/' + name + '/graph.bin'
        raw_dir = './openhgnn/dataset'
        url = self._prefix + self._urls[name]
        if name == 'amazon':
            url = 'https://zhiguli.oss-cn-hangzhou.aliyuncs.com/amazon.zip'
            self.data_path = './openhgnn/dataset/amazon.zip'
        super(AcademicDataset, self).__init__(name=name,
                                        url=url,
                                        raw_dir=raw_dir,
                                        force_reload=force_reload,
                                        verbose=verbose)

    def download(self):
        # download raw data to local disk
        # path to store the file
        if os.path.exists(self.data_path):  # pragma: no cover
           pass
        else:
            file_path = os.path.join(self.raw_dir)
            # download file
            download(self.url, path=file_path)
        extract_archive(self.data_path, os.path.join(self.raw_dir, self.name))

    def process(self):
        # process raw data to graphs, labels, splitting masks
        if self.name == 'yelp4HGSL':
            target_ntype = self._target_ntype
            canonical_etypes = self._canonical_etypes
            
            with open(self.raw_path + '/node_features.pkl', 'rb') as f:
                features = pickle.load(f)
            with open(self.raw_path + '/edges.pkl', 'rb') as f:
                edges = pickle.load(f)
            with open(self.raw_path + '/labels.pkl', 'rb') as f:
                labels = pickle.load(f)
            with open(self.raw_path + '/meta_data.pkl', 'rb') as f:
                meta_data = pickle.load(f)
            if scipy.sparse.issparse(features):
                features = features.todense()
            
            '''Load pretrained mp_embedding'''
            mp_emb_dict = {}
            mp_list = ['bub', 'bsb', 'bublb', 'bubsb']
            for mp in mp_list:
                f_name = self.raw_path + '/' + mp + '_emb.pkl'
                with open(f_name, 'rb') as f:
                    z = pickle.load(f)
                    zero_lines = np.nonzero(np.sum(z, 1) == 0)
                    if len(zero_lines) > 0:
                        # raise ValueError('{} zero lines in {}s!\nZero lines:{}'.format(len(zero_lines), mode, zero_lines))
                        z[zero_lines, :] += 1e-8
                    mp_emb_dict[mp] = z
            
            num_nodes = edges['s-b'].shape[0]
            assert len(canonical_etypes) == len(edges)

            ntype_mask = dict()
            ntype_idmap = dict()
            ntypes = set()
            data_dict = {}

            # create dgl graph
            for etype in canonical_etypes:
                ntypes.add(etype[0])
                ntypes.add(etype[2])
            for ntype in ntypes:
                ntype_mask[ntype] = np.zeros(num_nodes, dtype=bool)
                ntype_idmap[ntype] = np.full(num_nodes, -1, dtype=int)
            for etype in canonical_etypes:
                src_nodes = edges[etype[1]].nonzero()[0]
                dst_nodes = edges[etype[1]].nonzero()[1]
                src_ntype = etype[0]
                dst_ntype = etype[2]
                ntype_mask[src_ntype][src_nodes] = True
                ntype_mask[dst_ntype][dst_nodes] = True
            for ntype in ntypes:
                ntype_idx = ntype_mask[ntype].nonzero()[0]
                ntype_idmap[ntype][ntype_idx] = np.arange(ntype_idx.size)
            for etype in canonical_etypes:
                src_nodes = edges[etype[1]].nonzero()[0]
                dst_nodes = edges[etype[1]].nonzero()[1]
                src_ntype = etype[0]
                dst_ntype = etype[2]
                data_dict[etype] = \
                    (th.from_numpy(ntype_idmap[src_ntype][src_nodes]).type(th.int64),
                    th.from_numpy(ntype_idmap[dst_ntype][dst_nodes]).type(th.int64))
            g = dgl.heterograph(data_dict)

            # split and label
            all_label = np.full(g.num_nodes(target_ntype), -1, dtype=int)
            for i, split in enumerate(['train', 'val', 'test']):
                node = np.array(labels[i])[:, 0]
                label = np.array(labels[i])[:, 1]
                all_label[node] = label
                g.nodes[target_ntype].data['{}_mask'.format(split)] = \
                    th.from_numpy(idx2mask(node, g.num_nodes(target_ntype))).type(th.bool)
            g.nodes[target_ntype].data['label'] = th.from_numpy(all_label).type(th.long)

            # node feature
            node_features = th.from_numpy(features).type(th.FloatTensor)
            for ntype in ntypes:
                idx = ntype_mask[ntype].nonzero()[0]
                g.nodes[ntype].data['h'] = node_features[idx]

            for ntype in ntypes:
                idx = ntype_mask[ntype].nonzero()[0]
                for mp in mp_list:
                    tmp_tensor = th.from_numpy(mp_emb_dict[mp][idx])
                    g.nodes[ntype].data[mp] = tmp_tensor

            self._g = g
            self._num_classes = len(th.unique(self._g.nodes[self._target_ntype].data['label']))
            self._in_dim = self._g.ndata['h'][self._target_ntype].shape[1]
        else:
            g, _ = load_graphs(self.g_path)
            self._g = g[0]

    def __getitem__(self, idx):
        # get one example by index
        assert idx == 0, "This dataset has only one graph"
        return self._g

    def __len__(self):
        # number of data examples
        return 1

    def save(self):
        # save processed data to directory `self.save_path`
        pass

    def load(self):
        # load processed data from directory `self.save_path`
        pass

    def has_cache(self):
        # check whether there are processed data in `self.save_path`
        pass


import dgl
from dgl.sampling import RandomWalkNeighborSampler
from ..utils import extract_metapaths


class HANSampler(dgl.dataloading.Sampler):
    """HANSampler.
    Sample blocks by node types and meta paths.

    Parameters
    ----------
    g : DGLHeteroGraph
        The graph.
    ntypes : list[str]
        List of center node types.
    meta_paths_dict: dict[str, list[etype]]
        Dict from meta path name to meta path.
    num_neighbors: int
        Number of neighbors to sample.
    """

    def __init__(self, g, seed_ntypes, meta_paths_dict, num_neighbors):
        self.output_device = None  # as_edge_prediction_sampler requires this attribute

        self.ntype_mp_name_sampler_dict = {}
        self.seed_ntypes = seed_ntypes
        self.ntype_meta_paths_dict = {}

        # build ntype_meta_paths_dict
        for ntype in self.seed_ntypes:
            self.ntype_meta_paths_dict[ntype] = {}
            for meta_path_name, meta_path in meta_paths_dict.items():
                # a meta path starts with this node type
                if meta_path[0][0] == ntype:
                    self.ntype_meta_paths_dict[ntype][meta_path_name] = meta_path
        for ntype, meta_paths_dict in self.ntype_meta_paths_dict.items():
            if len(meta_paths_dict) == 0:
                self.ntype_meta_paths_dict[ntype] = extract_metapaths(ntype, g.canonical_etypes)

        for ntype, meta_paths_dict in self.ntype_meta_paths_dict.items():
            self.ntype_mp_name_sampler_dict[ntype] = {}
            for meta_path_name, meta_path in meta_paths_dict.items():
                # note: random walk may get same route(same edge), which will be removed in the sampled graph.
                # So the sampled graph's edges may be less than num_random_walks(num_neighbors).
                self.ntype_mp_name_sampler_dict[ntype][meta_path_name] = RandomWalkNeighborSampler(G=g,
                                                                                                   num_traversals=1,
                                                                                                   termination_prob=0,
                                                                                                   num_random_walks=num_neighbors,
                                                                                                   num_neighbors=num_neighbors,
                                                                                                   metapath=meta_path)

    def sample(self, g, seeds, exclude_eids=None):  # exclude_eids is for compatibility with link prediction
        """sample method.

        Returns
        -------
        dict[str, dict[str, Tensor]]
            Input node ids. Dict from node type to dict from meta path name to node ids.
        dict[str, Tensor]
            Seeds. Dict from node type to node ids
        dict[str, dict[str, DGLBlock]]
            Sampled blocks. Dict from node type to dict from meta path name to sampled blocks.
        """
        input_nodes_dict = {}
        ntype_mp_name_block_dict = {}
        for ntype, nid in seeds.items():
            if len(nid) == 0:
                continue
            input_nodes_dict[ntype] = {}
            ntype_mp_name_block_dict[ntype] = {}
            for meta_path_name, sampler in self.ntype_mp_name_sampler_dict[ntype].items():
                frontier = sampler(nid)
                frontier = dgl.remove_self_loop(frontier)
                frontier.add_edges(nid.clone().detach(), nid.clone().detach())
                block = dgl.to_block(frontier, nid)
                ntype_mp_name_block_dict[ntype][meta_path_name] = block
                input_nodes_dict[ntype][meta_path_name] = block.srcdata[dgl.NID]
        return input_nodes_dict, seeds, ntype_mp_name_block_dict



import dgl
import torch
import torch.nn.functional as F
from tqdm import tqdm
from ..utils.sampler import get_node_data_loader
from ..models import build_model
from . import BaseFlow, register_flow
from ..utils import EarlyStopping, to_hetero_idx, to_homo_feature, to_homo_idx
import math
from ..models.DSSL import LogisticRegression

import numpy as np


@register_flow("DSSL_trainer")
class DSSL_trainer(BaseFlow):

    def __init__(self,args = None):
        super(DSSL_trainer, self).__init__(args)            # build task(dataset..)

        self.args.category = self.task.dataset.category
        self.category = self.args.category

        self.num_classes = self.task.dataset.num_classes

        if not hasattr(self.task.dataset, 'out_dim') or args.out_dim != self.num_classes:
            self.logger.info('[NC Specific] Modify the out_dim with num_classes')
            args.out_dim = self.num_classes
        self.args.out_node_type = [self.category]
        self.args.feature_dim = self.hg.ndata['feat'].shape[1]
        # print(self.args.feature_dim)

        self.model = build_model(self.model).build_model_from_args(self.args, self.hg).to(self.device)
        self.optimizer = self.candidate_optimizer[args.optimizer](self.model.parameters(),
                                                                  lr=args.lr, weight_decay=args.weight_decay)

        if args.rand_split == True or args.dataset in ['snap-patents','ogbn-proteins', 'wiki','Cora', 'PubMed','genius']:
            self.train_idx, self.val_idx, self.test_idx = self.task.get_rand_split()
        else:
            self.train_idx, self.val_idx, self.test_idx = self.task.get_split()

        self.split_idx_lst = [{'train':self.train_idx,'valid':self.val_idx,'test':self.test_idx}]
        self.pred_idx = getattr(self.task.dataset, 'pred_idx', None)

        self.labels = self.task.get_labels().to(self.device)
        self.num_nodes_dict = {ntype: self.hg.num_nodes(ntype) for ntype in self.hg.ntypes}
        self.to_homo_flag = getattr(self.model, 'to_homo_flag', False)

        self.sampled_neighborhoods = self.sample_neighborhood(self.task.dataset, self.device, args)

    def train(self):

        print('MODEL:', self.model)

        # print (split_idx_lst)
        import datetime

        time_now = datetime.datetime.now()
        print('start training')
        print(time_now)

        meanAcc = 0

        split_idx = self.split_idx_lst[0]
        self.model.reset_parameters()
        optimizer = torch.optim.AdamW(self.model.parameters(), lr=self.args.lr, weight_decay=self.args.weight_decay)
        best_val = float('-inf')
        loss_lst = []
        best_loss = float('inf')

        for epoch in range(self.args.epochs):  # args.epochs = 200
            # pre-training
            self.model.train()
            batch_size = self.args.batch_size  # args.batch_size = 1024
            perm = torch.randperm(self.task.dataset.g.num_nodes())  # 随机打乱后获得的数字序列
            epoch_loss = 0
            for batch in range(0, self.task.dataset.g.num_nodes(), batch_size):
                optimizer.zero_grad()
                online_embedding = self.model.online_encoder(self.task.dataset)
                target_embedding = self.model.target_encoder(self.task.dataset)
                batch_idx = perm[batch:batch + batch_size]  # 从图中随机取batch_size个结点的id
                batch_idx = batch_idx.to(self.device)
                batch_neighbor_index = self.sampled_neighborhoods[batch_idx]  # 每个随机取的原结点取出有neighbor_max的邻居结点
                batch_embedding = online_embedding[batch_idx].to(self.device)  # 随机选的batch_size个结点的online_embedding
                batch_embedding = F.normalize(batch_embedding, dim=-1, p=2)  # 归一化 尺寸为   batch_size*dim=【1024，64】

                # target_embedding 中拿出随机原结点的邻居节点
                batch_neighbor_embedding = [target_embedding[i, :].unsqueeze(0) for i in batch_neighbor_index]
                batch_neighbor_embedding = torch.cat(batch_neighbor_embedding, dim=0).to(self.device)
                batch_neighbor_embedding = F.normalize(batch_neighbor_embedding, dim=-1, p=2)
                # batch_size*neighbor_max*dim=【1024，5，64】

                main_loss, context_loss, entropy_loss, k_node = self.model(batch_embedding, batch_neighbor_embedding)
                tmp = F.one_hot(torch.argmax(k_node, dim=1), num_classes=self.args.cluster_num).type(
                    torch.FloatTensor).to(self.device)
                batch_sum = (torch.reshape(torch.sum(tmp, 0), (-1, 1)))
                if self.args.neg_alpha:  # args.neg_alpha = 0
                    batch_neg_neighbor_index = sampled_neg_neighborhoods[batch_idx]
                    batch_neighbor_embedding = [target_embedding[i, :].unsqueeze(0) for i in
                                                batch_neg_neighbor_index]
                    batch_neighbor_embedding = torch.cat(batch_neighbor_embedding, dim=0).to(device)
                    batch_neighbor_embedding = F.normalize(batch_neighbor_embedding, dim=-1, p=2)
                    main_neg_loss, tmp, tmp, tmp = model(batch_embedding, batch_neighbor_embedding)
                    loss = main_loss + args.gamma * (context_loss + entropy_loss) + main_neg_loss

                else:
                    loss = main_loss + self.args.gamma * (context_loss + entropy_loss)
                print("batch : {}, main_loss: {}, context_loss: {}, entropy_loss: {}".format(batch,
                                                                                             main_loss,
                                                                                             context_loss,
                                                                                             entropy_loss))
                loss.backward()
                optimizer.step()
                self.model.update_moving_average()
                epoch_loss = epoch_loss + loss
            if epoch % 1 == 0:
                self.model.eval()
                for batch in range(0, self.task.dataset.g.num_nodes(), batch_size):
                    online_embedding = self.model.online_encoder(self.task.dataset)
                    target_embedding = self.model.target_encoder(self.task.dataset)
                    batch_idx = perm[batch:batch + batch_size]
                    batch_idx = batch_idx.to(self.device)
                    batch_neighbor_index = self.sampled_neighborhoods[batch_idx]
                    batch_target_embedding = target_embedding[batch_idx].to(self.device)
                    batch_embedding = online_embedding[batch_idx].to(self.device)
                    batch_neighbor_embedding = [target_embedding[i, :].unsqueeze(0) for i in batch_neighbor_index]
                    batch_neighbor_embedding = torch.cat(batch_neighbor_embedding, dim=0).to(self.device)
                    main_loss, context_loss, entropy_loss, k_node = self.model(batch_embedding, batch_neighbor_embedding)
                    tmp = F.one_hot(torch.argmax(k_node, dim=1), num_classes=self.args.cluster_num).type(
                        torch.FloatTensor).to(
                        self.device)
                    if batch == 0:
                        cluster = torch.matmul(batch_embedding.t(), tmp)
                        batch_sum = (torch.reshape(torch.sum(tmp, 0), (-1, 1)))
                    else:
                        cluster += torch.matmul(batch_embedding.t(), tmp)
                        batch_sum += (torch.reshape(torch.sum(tmp, 0), (-1, 1)))
                cluster = F.normalize(cluster, dim=-1, p=2)
                self.model.update_cluster(cluster, batch_sum)
            print("epoch: {}, loss: {}".format(epoch, epoch_loss))

            if epoch % 10 == 0:
                time_epoch = datetime.datetime.now()
                print("-------------------------------------------------------------------------------")
                print('cost time')
                timetmp = time_epoch - time_now
                print(timetmp)
                train_acc, dev_acc, test_acc = self.evaluate(self.model, self.task, split_idx)
                print("train acc: {}, val acc: {}, test acc: {}".format(train_acc, dev_acc, test_acc))
                print("-------------------------------------------------------------------------------")

                if best_val < dev_acc:
                    best_val = dev_acc
                    besttest_acc = test_acc

        meanAcc += besttest_acc
        print("meanAcc: {}".format(meanAcc))

    def _full_train_step(self):
        pass

    def _mini_train_step(self):
        pass

    def _full_test_step(self):
        pass

    def _mini_test_step(self):
        pass

    def sample_neighborhood(self,dataset, device, args):
        neighbor_max = args.neighbor_max  # args.neighbor_max = 5
        adjacent_list = {}
        # print(dataset.g)
        # print(dataset.g.num_nodes())
        # print(dataset.g.edges())

        undirect_g = dataset.g
        undirect_g = dgl.add_self_loop(undirect_g)
        undirect_g = dgl.to_bidirected(undirect_g)

        edge_index = undirect_g.edges()

        # adjacent_matrix = g.adj().to_dense()

        for i in range(0, dataset.g.num_nodes()):
            # adjacent_list[i] = torch.tensor(tmp,dtype=torch.int64).to(device)
            adjacent_list[i] = edge_index[1][edge_index[0] == i].to(device)  # 邻接矩阵



        for i in range(0, dataset.g.num_nodes()):
            num_neighbor = adjacent_list[i].shape[0]  # 第i个结点的邻居数量
            if num_neighbor < neighbor_max:
                more = math.ceil(neighbor_max / num_neighbor)
                adjacent_list[i] = adjacent_list[i].repeat(1, more)[0, :neighbor_max]  # 邻居数量不够使用重复填充
            if num_neighbor > neighbor_max:
                perm = torch.randperm(num_neighbor)
                idx = perm[:neighbor_max]
                adjacent_list[i] = adjacent_list[i][idx]  # 邻居数量超过使用在原num_neighbor里随机取neighbor_max个
        adjacent_list = list(adjacent_list.values())
        adjacent_list = torch.vstack(adjacent_list)
        return adjacent_list.to(device)

    def evaluate(self, model, task, split_idx):
        self.model.eval()
        embedding = self.model.online_encoder(task.dataset)
        embedding = embedding.detach()
        emb_dim, num_class = embedding.shape[1], self.num_classes
        train_accs, dev_accs, test_accs = [], [], []

        for i in range(10):
            train_idx = np.array(split_idx['train'].cpu())
            valid_idx = np.array(split_idx['valid'].cpu())
            test_idx = np.array(split_idx['test'].cpu())

            task.labels = task.labels.type(torch.LongTensor).to(self.device)
            task.labels = task.labels.unsqueeze(1)
            train_label = task.labels[train_idx]
            valid_label = task.labels[valid_idx]
            test_label = task.labels[test_idx]

            classifier = LogisticRegression(emb_dim, num_class).to(self.device)
            optimizer_LR = torch.optim.AdamW(classifier.parameters(), lr=0.01, weight_decay=0.01)

            for epoch in range(100):  # epoch（LR） = 100
                classifier.train()
                logits, loss = classifier(embedding[train_idx, :], train_label.squeeze())
                # print ("finetune epoch: {}, finetune loss: {}".format(epoch, loss))
                optimizer_LR.zero_grad()
                loss.backward()
                optimizer_LR.step()

            train_logits, _ = classifier(embedding[train_idx, :], train_label.squeeze())
            dev_logits, _ = classifier(embedding[valid_idx, :], valid_label.squeeze())
            test_logits, _ = classifier(embedding[test_idx, :], test_label.squeeze())
            train_preds = torch.argmax(train_logits, dim=1)
            dev_preds = torch.argmax(dev_logits, dim=1)
            test_preds = torch.argmax(test_logits, dim=1)

            train_acc = (torch.sum(train_preds == train_label.squeeze()).float() / train_label.squeeze().shape[
                0]).detach().cpu().numpy()
            dev_acc = (torch.sum(dev_preds == valid_label.squeeze()).float() / valid_label.squeeze().shape[
                0]).detach().cpu().numpy()
            test_acc = (torch.sum(test_preds == test_label.squeeze()).float() /
                        test_label.squeeze().shape[0]).detach().cpu().numpy()

            train_accs.append(train_acc * 100)
            dev_accs.append(dev_acc * 100)
            test_accs.append(test_acc * 100)

        train_accs = np.stack(train_accs)
        dev_accs = np.stack(dev_accs)
        test_accs = np.stack(test_accs)

        train_acc, train_std = train_accs.mean(), train_accs.std()
        dev_acc, dev_std = dev_accs.mean(), dev_accs.std()
        test_acc, test_std = test_accs.mean(), test_accs.std()

        return train_acc, dev_acc, test_acc


from . import BaseTask, register_task
from ..dataset import build_dataset


@register_task("meirec")
class MeiREC(BaseTask):
    """Recommendation tasks."""

    def __init__(self, args):
        super().__init__()
        self.logger = args.logger
        self.n_dataset = args.dataset
        self.train_loader, self.test_loader = build_dataset(args.dataset,
                                                            args.task,
                                                            args.batch_num,
                                                            logger=self.logger)

    def evaluate(self, y_true, y_score):
        pass




import dgl
import numpy as np
import torch as th
from tqdm import tqdm
import torch
from openhgnn.models import build_model
from . import BaseFlow, register_flow
from ..utils import EarlyStopping
import scipy.sparse as sp
import torch.nn.functional as F
@register_flow("DiffMG_trainer")
class DiffMG_trainer(BaseFlow):
    """Recommendation flows."""

    def __init__(self, args=None):
        super(DiffMG_trainer, self).__init__(args)
        self.target_link = self.task.dataset.target_link
        self.args.out_node_type = self.task.dataset.out_ntypes
        self.args.out_dim = self.args.hidden_dim
        self.model = build_model(self.model).build_model_from_args(self.args, self.hg)
        self.reg_weight = 0.1
        self.metric = ['recall', 'ndcg']
        self.val_metric = 'recall'
        self.topk = 20
        self.patience = args.patience
        self.max_epoch = args.max_epoch
        self.num_neg = self.task.dataset.num_neg
        self.user_name = self.task.dataset.user_name
        self.item_name = self.task.dataset.item_name
        self.num_user = self.hg.num_nodes(self.user_name)
        self.num_item = self.hg.num_nodes(self.user_name)
        self.train_eid_dict = {
            etype: self.hg.edges(etype=etype, form='eid')
            for etype in self.hg.canonical_etypes}

    def preprocess(self):
        if self.args.model == 'DiffMG':
            self.train_hg = self.task.train_hg
            self.val_hg = self.task.val_hg
            self.test_hg = self.task.test_hg
            train_mask = self.train_hg.edges['ui'].data['train_mask'].squeeze()
            train_index = th.nonzero(train_mask).squeeze()
            train_edge = self.train_hg.find_edges(train_index, 'ui')
            train_graph = dgl.heterograph({('user', 'ui', 'item'): train_edge},
                                          {ntype: self.train_hg.number_of_nodes(ntype) for ntype in ['user', 'item']})
            self.train_hg = train_graph
            process(self)
            return

    def train(self):
        self.preprocess()
        train(self)

def find_pos(edges):
    return (edges.data['rating'] > 3)

def find_neg(edges):
    return (edges.data['rating'] < 4)

def find_pairs(target, g, hg,type,u_num):
    eid = g.edges(etype=type)
    if target == 'pos':
        find = hg.filter_edges(find_pos, edges=(eid[0], eid[1]), etype=type)
        pairs = hg.find_edges(find, etype=type)
    else:
        find = hg.filter_edges(find_neg, edges=(eid[0], eid[1]), etype=type)
        pairs = hg.find_edges(find, etype=type)
    pairs = [pairs[0],pairs[1]]
    pairs[1] += u_num
    return pairs

def preprocess_amazon(self):
    hg = self.hg
    np.random.seed(self.args.Amazon_preprocess_seed)
    offsets = {'i': hg.num_nodes('user'), 'b': hg.num_nodes('item') + hg.num_nodes('user')}
    offsets['c'] = offsets['b'] + hg.num_nodes('brand')
    offsets['v'] = offsets['c'] + hg.num_nodes('category')
    in_dims = [hg.num_nodes('user'), hg.num_nodes('item'), hg.num_nodes('brand'),hg.num_nodes('category'), hg.num_nodes(
        'view')]
    node_types = np.zeros((hg.num_nodes(),), dtype=np.int32)
    node_types[offsets['i']:offsets['b']] = 1
    node_types[offsets['b']:offsets['c']] = 2
    node_types[offsets['c']:offsets['v']] = 3
    node_types[offsets['v']:] = 4
    self.train_pos = find_pairs('pos', self.train_hg, self.hg, 'ui', self.hg.num_nodes('user'))
    self.train_neg = find_pairs('neg', self.train_hg, self.hg, 'ui', self.hg.num_nodes('user'))
    self.test_pos = find_pairs('pos', self.test_hg, self.hg, 'ui', self.hg.num_nodes('user'))
    self.test_neg = find_pairs('neg', self.test_hg, self.hg, 'ui', self.hg.num_nodes('user'))
    self.val_pos = find_pairs('pos', self.val_hg, self.hg, 'ui', self.hg.num_nodes('user'))
    self.val_neg = find_pairs('neg', self.val_hg, self.hg, 'ui', self.hg.num_nodes('user'))
    keep_mask = hg.edges['ui'].data['keep_mask'].squeeze()
    keep_index = th.nonzero(keep_mask).squeeze()
    ui_edge = hg.find_edges(keep_index, 'ui')
    ib_edge = hg.edges(etype = 'ib')
    ic_edge = hg.edges(etype = 'ic')
    iv_edge = hg.edges(etype = 'iv')
    edge_dict = {'i':[ui_edge[0],ui_edge[1]],'b':[ib_edge[0],ib_edge[1]],'c':[ic_edge[0],ic_edge[1]],'v':[iv_edge[0],iv_edge[1]]}
    i = 1
    adjs_offset = {}
    for k,v in edge_dict.items():
        adj_offset = np.zeros((node_types.shape[0], node_types.shape[0]), dtype=np.float32)
        adj_offset[v[0].tolist(), (v[1]+ offsets[k]).tolist() ] = 1
        adjs_offset[str(i)] = sp.coo_matrix(adj_offset)
        i = i+1
    self.node_types = node_types
    self.adjs_offset = adjs_offset
    # self.ndarry_dict = ndarry_dict
    self.in_dims = in_dims
def process(self):
    self.cstr_source = {  #* u-side
    "amazon" : [0, 8],
}
    self.cstr_target = {  #* i-side
    "amazon" : [1, 2, 4, 6, 8],
}
    if self.args.gpu > -1:
        print("use gpu---------------")
        torch.cuda.set_device(self.args.gpu)
        torch.cuda.manual_seed(self.args.Amazon_train_seed)
    else:
        print("use cpu_______________")
    np.random.seed(self.args.seed)
    torch.manual_seed(self.args.seed)

    preprocess_amazon(self)
    num_node_types = self.node_types.max() + 1
    node_types = torch.from_numpy(self.node_types)
    adjs_pt = []
    if '0' in self.adjs_offset:
            adjs_pt.append(sparse_mx_to_torch_sparse_tensor(
                normalize_sym(self.adjs_offset['0'] + sp.eye(self.adjs_offset['0'].shape[0], dtype=np.float32))))
    for i in range(1, int(max(self.adjs_offset.keys())) + 1):
            adjs_pt.append(sparse_mx_to_torch_sparse_tensor(
                normalize_row(self.adjs_offset[str(i)] + sp.eye(self.adjs_offset[str(i)].shape[0], dtype=np.float32))))
            adjs_pt.append(sparse_mx_to_torch_sparse_tensor(
                normalize_row(self.adjs_offset[str(i)].T + sp.eye(self.adjs_offset[str(i)].shape[0], dtype=np.float32))))
    adjs_pt.append(
            sparse_mx_to_torch_sparse_tensor(sp.eye(self.adjs_offset['1'].shape[0], dtype=np.float32).tocoo()))
    adjs_pt.append(torch.sparse.FloatTensor(size=self.adjs_offset['1'].shape))
    print("Loading {} adjs...".format(len(adjs_pt)))
    self.serach_adjs_pt = []
    if '0' in self.adjs_offset:
        self.serach_adjs_pt.append(sparse_mx_to_torch_sparse_tensor(normalize_sym(self.adjs_offset['0'] + sp.eye(self.adjs_offset['0'].shape[0], dtype=np.float32))))
    for i in range(1, int(max(self.adjs_offset.keys())) + 1):
        self.serach_adjs_pt.append(sparse_mx_to_torch_sparse_tensor(normalize_row(self.adjs_offset[str(i)] + sp.eye(self.adjs_offset[str(i)].shape[0], dtype=np.float32))))
        self.serach_adjs_pt.append(sparse_mx_to_torch_sparse_tensor(normalize_row(self.adjs_offset[str(i)].T + sp.eye(self.adjs_offset[str(i)].shape[0], dtype=np.float32))))
    self.serach_adjs_pt.append(sparse_mx_to_torch_sparse_tensor(sp.eye(self.adjs_offset['1'].shape[0], dtype=np.float32).tocoo()))
    self.serach_adjs_pt.append(torch.sparse.FloatTensor(size=self.adjs_offset['1'].shape))
    node_feats = []
    for k in range(num_node_types):
        i = torch.stack((torch.arange(self.in_dims[k], dtype=torch.long), torch.arange(self.in_dims[k], dtype=torch.long)))
        v = torch.ones(self.in_dims[k])
        node_feats.append(torch.sparse.FloatTensor(i, v, torch.Size([self.in_dims[k], self.in_dims[k]])))
    self.node_feats = node_feats
    self.adjs_pt = adjs_pt

    search(self)
    steps_s = [len(meta) for meta in self.archs[self.args.dataset]["source"][0]]
    steps_t = [len(meta) for meta in self.archs[self.args.dataset]["target"][0]]
    model_s = self.model(self.in_dims, self.args.attn_dim, steps_s, dropout=self.args.dropout)
    model_t = self.model(self.in_dims, self.args.attn_dim, steps_t, dropout=self.args.dropout)
    optimizer = torch.optim.Adam(
        list(model_s.parameters()) + list(model_t.parameters()),
        lr=self.args.lr,
        weight_decay=self.args.wd
    )
    self.steps_s = steps_s
    self.steps_t = steps_t
    self.model_s = model_s
    self.model_t = model_t
    self.optimizer = optimizer
def train1(self):
    self.model_s.train()
    self.model_t.train()
    self.optimizer.zero_grad()
    out_s = self.model_s(self.node_feats, self.node_types, self.adjs_pt, self.archs[self.args.dataset]["source"][0],
                    self.archs[self.args.dataset]["source"][1], self.args.gpu)
    out_t = self.model_t(self.node_feats, self.node_types, self.adjs_pt, self.archs[self.args.dataset]["target"][0],
                    self.archs[self.args.dataset]["target"][1], self.args.gpu)
    loss = - torch.mean(F.logsigmoid(torch.mul(out_s[self.train_pos[0].tolist()], out_t[self.train_pos[1].tolist()]).sum(dim=-1)) + \
                        F.logsigmoid(- torch.mul(out_s[self.train_neg[0].tolist()], out_t[self.train_neg[1].tolist()]).sum(dim=-1)))
    loss.backward()
    self.optimizer.step()
    return loss.item()

def infer(self):
    import torch.nn.functional as F
    from sklearn.metrics import roc_auc_score
    self.model_s.eval()
    self.model_t.eval()
    with torch.no_grad():
        out_s = self.model_s(self.node_feats, self.node_types, self.adjs_pt, self.archs[self.args.dataset]["source"][0],
                        self.archs[self.args.dataset]["source"][1], self.args.gpu)
        out_t = self.model_t(self.node_feats, self.node_types, self.adjs_pt, self.archs[self.args.dataset]["target"][0],
                        self.archs[self.args.dataset]["target"][1], self.args.gpu)

    # * validation performance
    pos_val_prod = torch.mul(out_s[self.val_pos[0].tolist()], out_t[self.val_pos[1].tolist()]).sum(dim=-1)
    neg_val_prod = torch.mul(out_s[self.val_neg[0].tolist()], out_t[self.val_neg[1].tolist()]).sum(dim=-1)
    loss = - torch.mean(F.logsigmoid(pos_val_prod) + F.logsigmoid(- neg_val_prod))
    y_true_val = np.zeros((len(self.val_pos[0]) + len(self.val_neg[0])), dtype=np.long)
    y_true_val[:len(self.val_pos[0])] = 1
    y_pred_val = np.concatenate(
        (torch.sigmoid(pos_val_prod).cpu().numpy(), torch.sigmoid(neg_val_prod).cpu().numpy()))
    auc_val = roc_auc_score(y_true_val, y_pred_val)
    # * test performance
    pos_test_prod = torch.mul(out_s[self.test_pos[0].tolist()], out_t[self.test_pos[1].tolist()]).sum(dim=-1)
    neg_test_prod = torch.mul(out_s[self.test_neg[0].tolist()], out_t[self.test_neg[1].tolist()]).sum(dim=-1)

    y_true_test = np.zeros((len(self.test_pos[0]) + len(self.test_neg[0])), dtype=np.long)
    y_true_test[:len(self.test_pos[0])] = 1
    y_pred_test = np.concatenate(
        (torch.sigmoid(pos_test_prod).cpu().numpy(), torch.sigmoid(neg_test_prod).cpu().numpy()))
    auc_test = roc_auc_score(y_true_test, y_pred_test)
    return loss.item(), auc_val, auc_test

def train(self):
    import logging
    best_val = None
    final = None
    anchor = None
    for epoch in range(self.max_epoch):
        train_loss = train1(self)
        val_loss, auc_val, auc_test = infer(self)
        logging.info(
            "Epoch {}; Train err {}; Val err {}; Val auc {}".format(epoch + 1, train_loss, val_loss, auc_val))
        if best_val is None or auc_val > best_val:
            best_val = auc_val
            final = auc_test
            anchor = epoch + 1
    logging.info("Best val auc {} at epoch {}; Test auc {}".format(best_val, anchor, final))

def search(self):
    # torch.cuda.set_device(args.gpu)
    import logging
    np.random.seed(self.args.Amazon_search_seed)
    torch.manual_seed(self.args.Amazon_search_seed)

    model_s = self.args.search_model(self.in_dims, self.args.attn_dim, len( self.serach_adjs_pt), [self.args.search_steps_s], self.cstr_source[self.args.dataset])
    model_t = self.args.search_model(self.in_dims, self.args.attn_dim, len( self.serach_adjs_pt), [self.args.search_steps_t], self.cstr_target[self.args.dataset])
    optimizer_w = torch.optim.Adam(
        list(model_s.parameters()) + list(model_t.parameters()),
        lr=self.args.search_lr,
        weight_decay=self.args.search_wd
    )

    optimizer_a = torch.optim.Adam(
        model_s.alphas() + model_t.alphas(),
        lr=self.args.search_alr
    )

    eps = self.args.search_eps
    for epoch in range(self.args.search_epochs):
        min_error = 100
        train_error, val_error = serach_train(self,model_s, model_t, optimizer_w, optimizer_a, eps)
        logging.info(
            "Epoch {}; Train err {}; Val err {}; Source arch {}; Target arch {}".format(epoch + 1, train_error,
                                                                                        val_error, model_s.parse(),
                                                                                        model_t.parse()))
        if val_error < min_error:
            self.archs = {
                "amazon": {
                    "source": model_s.parse(),
                    "target": model_t.parse()
                },
            }
            min_error = val_error
        eps = eps * self.args.search_decay

def serach_train(self, model_s, model_t, optimizer_w,
          optimizer_a, eps):

    idxes_seq_s, idxes_res_s = model_s.sample(eps)
    idxes_seq_t, idxes_res_t = model_t.sample(eps)

    optimizer_w.zero_grad()
    out_s = model_s(self.node_feats, self.node_types, self.serach_adjs_pt, idxes_seq_s, idxes_res_s)
    out_t = model_t(self.node_feats, self.node_types,  self.serach_adjs_pt, idxes_seq_t, idxes_res_t)
    loss_w = - torch.mean(F.logsigmoid(torch.mul(out_s[self.train_pos[0].tolist()], out_t[self.train_pos[1].tolist()]).sum(dim=-1)) + \
                          F.logsigmoid(- torch.mul(out_s[self.train_neg[0].tolist()], out_t[self.train_neg[1].tolist()]).sum(dim=-1)))
    loss_w.backward()
    optimizer_w.step()

    optimizer_a.zero_grad()
    out_s = model_s(self.node_feats, self.node_types,  self.serach_adjs_pt, idxes_seq_s, idxes_res_s)
    out_t = model_t(self.node_feats, self.node_types,  self.serach_adjs_pt, idxes_seq_t, idxes_res_t)
    loss_a = - torch.mean(F.logsigmoid(torch.mul(out_s[self.val_pos[0].tolist()], out_t[self.val_pos[1].tolist()]).sum(dim=-1)) + \
                          F.logsigmoid(- torch.mul(out_s[self.val_neg[0].tolist()], out_t[self.val_neg[1].tolist()]).sum(dim=-1)))
    loss_a.backward()
    optimizer_a.step()

    return loss_w.item(), loss_a.item()

def normalize_sym(adj):
    """Symmetrically normalize adjacency matrix."""
    rowsum = np.array(adj.sum(1))
    d_inv_sqrt = np.power(rowsum, -0.5).flatten()
    d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.
    d_mat_inv_sqrt = sp.diags(d_inv_sqrt)
    return adj.dot(d_mat_inv_sqrt).transpose().dot(d_mat_inv_sqrt).tocoo()
def normalize_row(mx):
    """Row-normalize sparse matrix"""
    rowsum = np.array(mx.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    mx = r_mat_inv.dot(mx)
    return mx.tocoo()
def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    indices = torch.from_numpy(
    np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse.FloatTensor(indices, values, shape)


import torch as th
from tqdm import tqdm
from . import BaseFlow, register_flow
from ..models import build_model
from ..utils import EarlyStopping
from ..sampler.TransX_sampler import TransX_Sampler


@register_flow("TransX_trainer")
class TransXTrainer(BaseFlow):
    """TransX flows."""

    def __init__(self, args):
        super(TransXTrainer, self).__init__(args)

        self.args = args
        self.model_name = args.model
        self.device = args.device
        self.batch_size = args.batch_size
        self.neg_size = args.neg_size
        self.score_size = args.batch_size * (args.neg_size * 2 + 1)
        self.max_epoch = args.max_epoch
        self.margin = args.margin

        self.train_hg = self.task.get_train()
        self.model = build_model(self.model).build_model_from_args(self.args, self.hg)
        self.model = self.model.to(self.device)

        self.optimizer = self.candidate_optimizer[args.optimizer](self.model.parameters(),
                                                                  lr=args.lr, weight_decay=args.weight_decay)

        self.stopper = EarlyStopping(args.patience, self._checkpoint)
        self.task.ScorePredictor = self.model.forward  # new score prdictor here

        self.num_nodes = getattr(self.task.dataset, 'num_nodes', self.hg.num_nodes())
        self.num_rels = getattr(self.task.dataset, 'num_rels', self.hg.num_edges())

    def preprocess(self):
        self.load_from_pretrained()
        self.train_hg.to(self.device)
        self.train_sampler = TransX_Sampler(self.train_hg, self.args)
        self.node_range = th.arange(0, self.num_nodes).to(self.device)
        self.rel_range = th.arange(0, self.num_rels).to(self.device)

        if self.args.score_fn == 'transr':  # load transe data when training transr
            transe_state_dict = th.load(self.stopper.save_path.replace("TransR", "TransE"))
            self.model.n_emb.weight.data = transe_state_dict['n_emb.weight']
            self.model.r_emb.weight.data = transe_state_dict['r_emb.weight']
            print("load")

    def train(self):
        self.preprocess()
        epoch = self._train()
        self.stopper.load_model(self.model)
        if self.args.test_flag:
            test_matrix = self._test()
            return dict(metric=test_matrix, epoch=epoch)
        if self.args.prediction_flag:
            return self._pred_step()

    def _train(self):
        batch_num = self.train_sampler.batch_num
        for epoch in range(self.max_epoch):
            self.logger.info(f"[Train Info] epoch {epoch:03d}")
            self.model.train()
            loss_sum = 0
            iter_range = tqdm(range(batch_num), ncols=100)
            for iter in iter_range:
                self.optimizer.zero_grad()
                pos_g = self.train_sampler.get_pos()
                neg_g = self.train_sampler.get_neg()
                h_emb, r_emb, t_emb = th.cat((pos_g[0], neg_g[0]), -1), th.cat((pos_g[1], neg_g[1]), -1), th.cat(
                    (pos_g[2], neg_g[2]), -1)
                loss = self.loss_calculation(h_emb, r_emb, t_emb)
                loss.backward()
                self.optimizer.step()
                loss_sum += loss.item()

            self.logger.info(f"[Train Info] epoch {epoch:03d} loss: {loss_sum}")

            if epoch % self.evaluate_interval == 0:
                val_metric = self._test_step('valid')
                self.logger.info("[Evaluation metric] " + str(val_metric))  # out test result
                early_stop = self.stopper.loss_step(val_metric['valid']['MR'], self.model)  # less is better
                if early_stop:
                    self.logger.train_info(f'Early Stop!\tEpoch:{epoch:03d}.')
                    break
        if self.max_epoch == 0:
            return 0
        else:
            return epoch

    def loss_calculation(self, h_emb, r_emb, t_emb):
        score = self.task.ScorePredictor(h_emb, r_emb, t_emb)

        if score.size(dim=0) == self.score_size:
            batch_size = self.batch_size
        else:  # last batch of a round
            batch_size = score.size(dim=0) // (self.neg_size * 2 + 1)

        p_score = score[:batch_size]
        p_score = p_score.view(batch_size, 1)

        n_score = score[batch_size:]
        n_score_split = th.chunk(n_score, 2, dim=0)
        n_score_tail = n_score_split[0].view(batch_size, self.neg_size)  # change tail
        n_score_head = n_score_split[1].view(batch_size, self.neg_size)  # change head
        n_score = th.cat((n_score_head, n_score_tail), dim=1)

        loss = th.clamp(p_score - n_score + self.margin, min=0.0).mean()
        return loss

    def _test(self):
        test_metric = self._test_step('test')
        self.logger.info("[Test Info] " + str(test_metric))  # out test result
        return test_metric  # dict

    def _test_step(self, mode):
        self.model.eval()
        with th.no_grad():
            n_emb = th.arange(self.num_nodes)
            r_emb_pre = th.arange(self.num_rels)
            # n_emb, r_emb_pre, _ = self.model(self.node_range, self.rel_range, th.tensor(0))
            r_emb = {}
            for i in range(self.num_rels):
                r_emb[i] = r_emb_pre[i]
            return {mode: self.task.evaluate(n_emb, r_emb, mode)}

    def _pred_step(self):
        self.model.eval()
        with th.no_grad():
            return self.task.tranX_predict()
